# gSpan is a subgraph mining algorithm that uses DFS and DFS codes to mine subgraphs
#
# **Importing this algorithm into a python program**
#
#             from PAMI.subgraphMining.basic import gspan as alg
#
#             obj = alg.GSpan(iFile, minSupport)
#
#             obj.mine()
#
#             obj.run()
#
#             frequentGraphs = obj.getFrequentSubgraphs()
#
#             memUSS = obj.getMemoryUSS()
#
#             obj.save(oFile)
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
class ExtendedEdge:
    def __init__(self, v1, v2, vLabel1, vLabel2, edgeLabel):
        self.v1 = v1
        self.v2 = v2
        self.vLabel1 = vLabel1
        self.vLabel2 = vLabel2
        self.edgeLabel = edgeLabel
        self.hashcode = (1 + v1) * 100 + (1 + v2) * 50 + (1 + vLabel1) * 30 + (1 + vLabel2) * 20 + (1 + edgeLabel)

    def smallerThan(self, that):
        if that is None:
            return True

        x1, x2, y1, y2 = self.v1, self.v2, that.v1, that.v2

        if self.pairSmallerThan(x1, x2, y1, y2):
            return True
        elif x1 == y1 and x2 == y2:
            return (self.vLabel1 < that.vLabel1 or
                    (self.vLabel1 == that.vLabel1 and self.vLabel2 < that.vLabel2) or
                    (self.vLabel1 == that.vLabel1 and self.vLabel2 == that.vLabel2 and
                     self.edgeLabel < that.edgeLabel))
        else:
            return False

  
    def smallerThanOriginal(self, that):
        if that is None:
            return True

        x1, x2, y1, y2 = self.v1, self.v2, that.v1, that.v2

        if self.pairSmallerThan(x1, x2, y1, y2):
            return True
        elif x1 == y1 and x2 == y2:
            return (self.vLabel1 < that.vLabel1 or
                    (self.vLabel1 == that.vLabel1 and self.edgeLabel < that.edgeLabel) or
                    (self.vLabel1 == that.vLabel1 and self.edgeLabel == that.edgeLabel and
                     self.vLabel2 < that.vLabel2))
        else:
            return False

    def pairSmallerThan(self, x1, x2, y1, y2):
        xForward = x1 < x2
        yForward = y1 < y2

        if xForward and yForward:
            return x2 < y2 or (x2 == y2 and x1 > y1)
        elif not xForward and not yForward:
            return x1 < y1 or (x1 == y1 and x2 < y2)
        elif xForward:
            return x2 <= y1
        else:
            return x1 < y2

    def __hash__(self):
        return self.hashcode

    def __eq__(self, other):
        if not isinstance(other, ExtendedEdge):
            return False
        return (self.v1 == other.v1 and self.v2 == other.v2 and
                self.vLabel1 == other.vLabel1 and self.vLabel2 == other.vLabel2 and
                self.edgeLabel == other.edgeLabel)

    def __repr__(self):
        return f"<{self.v1},{self.v2},{self.vLabel1},{self.vLabel2},{self.edgeLabel}>"

    def getV1(self):
        return self.v1
    
    def getV2(self):
        return self.v2
    
    def getVLabel1(self):
        return self.vLabel1
    
    def getVLabel2(self):
        return self.vLabel2
    
    def getEdgeLabel(self):
        return self.edgeLabel



# plotLineGraphFromDictionary is used to convert the given dictionary into plotLineGraph.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#     from PAMI.extras.graph import plotLineGraphFromDictionary as plt
#
#     obj = plt.plotLineGraphFromDictionary(idict, 100, 0, " ")
#
#     obj.save()
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import matplotlib.pyplot as plt

class plotLineGraphFromDictionary:
    """
    This class plot graph of input data

    :Attributes:

    :param data : dict: store input data as dict

    :Methods:

        plotLineGraph()
            draw line graph of input data. input data's key is x and value is y.

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.graph import plotLineGraphFromDictionary as plt

            obj = plt.plotLineGraphFromDictionary(idict, 100, 0, " ")

            obj.save()
    """
    def __init__(self, data: dict, end: int=100, start: int=0, title: str='', xlabel: str='', ylabel: str='') -> None:

        """
        draw line graph. Plot the input data key as x and value as y

        :param end: end of graph to plot
        :type end: int
        :param start: start fo graph to plot
        :type start: int
        :param title: title of graph
        :type title: str
        :param xlabel: xlabel of graph
        :type xlabel: str
        :param ylabel: ylabel of grapth
        :type ylabel: str
        :return: None
        """
        end = int(len(data) * end / 100)
        start = int(len(data) * start / 100)
        x = list(range(len(data)))
        y = list(tuple(data.values())[start:end])
        fig, ax = plt.subplots()
        ax.plot(x, y, marker='.')
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

if __name__ == '__main__':
    obj = plotLineGraphFromDictionary(sys.argv[1])
    obj.plotLineGraphFromDictionary(sys.argv[2])


#  CPFPMiner algorithm is used to discover the closed periodic frequent patterns in temporal databases.
#  It uses depth-first search.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.periodicFrequentPattern.closed import CPFPMiner as alg
#
#             obj = alg.CPFPMiner("../basic/sampleTDB.txt", "2", "6")
#
#             obj.startMine()
#
#             periodicFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(periodicFrequentPatterns))
#
#             obj.save("patterns")
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#



__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.periodicFrequentPattern.basic import abstract as _ab
import pandas as pd
from deprecated import deprecated

from PAMI.periodicFrequentPattern.closed import abstract as _ab

class CPFPMiner(_ab._periodicFrequentPatterns):
    """
    About this algorithm
    ====================

    :Description:   CPFPMiner algorithm is used to discover the closed periodic frequent patterns in temporal databases.
                    It uses depth-first search.

    :Reference:   P. Likhitha et al., "Discovering Closed Periodic-Frequent Patterns in Very Large Temporal Databases"
                  2020 IEEE International Conference on Big Data (Big Data), 2020, https://ieeexplore.ieee.org/document/9378215

    :param  iFile: str :
                   Name of the Input file to mine complete set of periodic frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of periodic frequent pattern's
    :param  minSup: float:
                   Controls the minimum number of transactions in which every item must appear in a database.
    :param  maxPer: float:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        iFile : str
            Input file name or path of the input file
        oFile : str
            Name of the output file or path of the input file
        minSup: int or float or str
            The user can specify minSup either in count or proportion of database size.
            If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
            Otherwise, it will be treated as float.
            Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
        maxPer: int or float or str
            The user can specify maxPer either in count or proportion of database size.
            If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program

    :Methods:

        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of frequent patterns will be loaded in to an output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function


    Execution methods
    =================


    **Terminal command**


    .. code-block:: console

       Format:

       (.venv) $  python3 CPFPMiner.py <inputFile> <outputFile> <minSup> <maxPer>

       Example:

       (.venv) $ python3 CPFPMiner.py sampleTDB.txt patterns.txt 0.3 0.4

     .. note:: minSup will be considered in percentage of database transactions
        
        
    **Calling from a python program**

    .. code-block:: python

                    from PAMI.periodicFrequentPattern.closed import CPFPMiner as alg
        
                    obj = alg.CPFPMiner("../basic/sampleTDB.txt", "2", "6")
        
                    obj.startMine()
        
                    periodicFrequentPatterns = obj.getPatterns()
        
                    print("Total number of Frequent Patterns:", len(periodicFrequentPatterns))
        
                    obj.save("patterns")
        
                    Df = obj.getPatternsAsDataFrame()
        
                    memUSS = obj.getMemoryUSS()
        
                    print("Total Memory in USS:", memUSS)
        
                    memRSS = obj.getMemoryRSS()
        
                    print("Total Memory in RSS", memRSS)
        
                    run = obj.getRuntime()
        
                    print("Total ExecutionTime in seconds:", run)
        
    **Credits:**
    ------------------
    The complete program was written by  P.Likhitha under the supervision of Professor Rage Uday Kiran.
    """

    _minSup = float()
    _maxPer = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _transaction = []
    _hashing = {}
    _mapSupport = {}
    _itemSetCount = 0
    _maxItemId = 0
    _tableSize = 10000
    _tidList = {}
    _lno = 0

    def __init__(self, iFile, minSup, maxPer, sep='\t'):
        super().__init__(iFile, minSup, maxPer, sep)
        self._finalPatterns = {}
    
    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._lno * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._lno * value)
            else:
                value = int(value)
        return value

    def _scanDatabase(self):
        """
        To scan the database and extracts the 1-length periodic-frequent items

        :return:   Returns the 1-length periodic-frequent items
        """
        Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            ts, data = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                Database.append(tr)

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()
        self._tidList = {}
        self._mapSupport = {}
        for line in Database:
            self._lno += 1
            s = line
            n = int(s[0])
            for i in range(1, len(s)):
                si = s[i]
                if self._mapSupport.get(si) is None:
                    self._mapSupport[si] = [1, abs(0 - n), n]
                    self._tidList[si] = [n]
                else:
                    self._mapSupport[si][0] += 1
                    self._mapSupport[si][1] = max(self._mapSupport[si][1], abs(n - self._mapSupport[si][2]))
                    self._mapSupport[si][2] = n
                    self._tidList[si].append(n)
        for x, y in self._mapSupport.items():
            self._mapSupport[x][1] = max(self._mapSupport[x][1], abs(self._lno - self._mapSupport[x][2]))
        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        self._mapSupport = {k: [v[0], v[1]] for k, v in self._mapSupport.items() if
                           v[0] >= self._minSup and v[1] <= self._maxPer}
        periodicFrequentItems = {}
        self._tidList = {k: v for k, v in self._tidList.items() if k in self._mapSupport}
        for x, y in self._tidList.items():
            t1 = 0
            for i in y:
                t1 += i
            periodicFrequentItems[x] = t1
        periodicFrequentItems = [key for key, value in sorted(periodicFrequentItems.items(), key=lambda x: x[1])]
        return periodicFrequentItems

    def _calculate(self, tidSet):
        """
        To calculate the weight if pattern based on the respective timeStamps

        :param tidSet: timeStamps of the pattern
        :return: the calculated weight of the timeStamps
        """
        hashcode = 0
        for i in tidSet:
            hashcode += i
        if hashcode < 0:
            hashcode = abs(0 - hashcode)
        return hashcode % self._tableSize

    def _contains(self, itemSet, val, hashcode):
        """
        To check if the key(hashcode) is in dictionary(hashing) variable

        :param itemSet: generated periodic-frequent itemSet
        :param val: support and periodicity of itemSet
        :param hashcode: the key generated in calculate() method for every itemSet

        :return: true if itemSet with same support present in dictionary(hashing) or else returns false
        """
        if self._hashing.get(hashcode) is None:
            return False
        for i in self._hashing[hashcode]:
            itemSetX = i
            if val[0] == self._hashing[hashcode][itemSetX][0] and set(itemSetX).issuperset(itemSet):
                return True
        return False

    def _getPeriodAndSupport(self, timeStamps):
        """
        Calculates the periodicity and support of timeStamps

        :param timeStamps: timeStamps of itemSet
        :return: periodicity and support
        """
        timeStamps.sort()
        cur = 0
        per = 0
        sup = 0
        for j in range(len(timeStamps)):
            per = max(per, timeStamps[j] - cur)
            if per > self._maxPer:
                return [0, 0]
            cur = timeStamps[j]
            sup += 1
        per = max(per, self._lno - cur)
        return [sup, per]

    def _save(self, prefix, suffix, tidSetX):
        """
        Saves the generated pattern which satisfies the closed property
        Parameters:
        -----------
            prefix: the prefix part of itemSet
            suffix: the suffix part of itemSet
            tidSetX: the timeStamps of the generated itemSet

        Returns:
        --------
            saves the closed periodic-frequent pattern

        """
        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        prefix = list(set(prefix))
        prefix.sort()
        val = self._getPeriodAndSupport(tidSetX)
        if val[0] >= self._minSup and val[1] <= self._maxPer:
            hashcode = self._calculate(tidSetX)
            if self._contains(prefix, val, hashcode) is False:
                self._itemSetCount += 1
                sample = str()
                for i in prefix:
                    sample = sample + i + " "
                self._finalPatterns[sample] = val
            if hashcode not in self._hashing:
                self._hashing[hashcode] = {tuple(prefix): val}
            else:
                self._hashing[hashcode][tuple(prefix)] = val

    def _processEquivalenceClass(self, prefix, itemSets, tidSets):
        """
        identifies and saves closed periodic patterns of length more than 2 in a dataset, by processing equivalence classes of item sets that satisfy a minimum support condition.
        :param prefix: Prefix class of an itemSet
        :param itemSets: suffix items in periodicFrequentItems that satisfies the minSup condition
        :param tidSets: timeStamps of items in itemSets respectively
        :return:  closed periodic patterns with length more than 2
        """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidList = tidSets[0]
            self._save(prefix, [i], tidList)
            return
        if len(itemSets) == 2:
            itemI = itemSets[0]
            tidSetI = tidSets[0]
            itemJ = itemSets[1]
            tidSetJ = tidSets[1]
            y1 = list(set(tidSetI).intersection(tidSetJ))
            if len(y1) >= self._minSup:
                suffix = []
                suffix += [itemI, itemJ]
                suffix = list(set(suffix))
                self._save(prefix, suffix, y1)
            if len(y1) != len(tidSetI):
                self._save(prefix, [itemI], tidSetI)
            if len(y1) != len(tidSetJ):
                self._save(prefix, [itemJ], tidSetJ)
            return
        for i in range(len(itemSets)):
            itemX = itemSets[i]
            if itemX is None:
                continue
            tidSetX = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetX = [itemX]
            for j in range(i + 1, len(itemSets)):
                itemJ = itemSets[j]
                if itemJ is None:
                    continue
                tidSetJ = tidSets[j]
                y = list(set(tidSetX).intersection(tidSetJ))
                if len(y) < self._minSup:
                    continue
                if len(tidSetX) == len(tidSetJ) and len(y) == len(tidSetX):
                    itemSets.insert(j, None)
                    tidSets.insert(j, None)
                    itemSetX.append(itemJ)
                elif len(tidSetX) < len(tidSetJ) and len(y) == len(tidSetX):
                    itemSetX.append(itemJ)
                elif len(tidSetX) > len(tidSetJ) and len(y) == len(tidSetJ):
                    itemSets.insert(j, None)
                    tidSets.insert(j, None)
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
                else:
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            if len(classItemSets) > 0:
                newPrefix = list(set(itemSetX)) + prefix
                self._processEquivalenceClass(newPrefix, classItemSets, classTidSets)
            self._save(prefix, list(set(itemSetX)), tidSetX)

    @deprecated("It is recommended to use mine() instead of startMine() for mining process")
    def startMine(self):
        """
        Mining process will start from here
        """
        self._startTime = _ab._time.time()
        self._finalPatterns = {}
        self._hashing = {}
        periodicFrequentItems = self._scanDatabase()
        for i in range(len(periodicFrequentItems)):
            itemX = periodicFrequentItems[i]
            if itemX is None:
                continue
            tidSetX = self._tidList[itemX]
            itemSetX = [itemX]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(periodicFrequentItems)):
                itemJ = periodicFrequentItems[j]
                if itemJ is None:
                    continue
                tidSetJ = self._tidList[itemJ]
                y1 = list(set(tidSetX).intersection(tidSetJ))
                if len(y1) < self._minSup:
                    continue
                if len(tidSetX) == len(tidSetJ) and len(y1) is len(tidSetX):
                    periodicFrequentItems.insert(j, None)
                    itemSetX.append(itemJ)
                elif len(tidSetX) < len(tidSetJ) and len(y1) is len(tidSetX):
                    itemSetX.append(itemJ)
                elif len(tidSetX) > len(tidSetJ) and len(y1) is len(tidSetJ):
                    periodicFrequentItems.insert(j, None)
                    itemSets.append(itemJ)
                    tidSets.append(y1)
                else:

                    itemSets.append(itemJ)
                    tidSets.append(y1)
            if len(itemSets) > 0:
                self._processEquivalenceClass(itemSetX, itemSets, tidSets)
            self._save([], itemSetX, tidSetX)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Closed periodic frequent patterns were generated successfully using CPFPMiner algorithm ")

    def Mine(self):
        """
        Mining process will start from here
        """
        self._startTime = _ab._time.time()
        self._finalPatterns = {}
        self._hashing = {}
        periodicFrequentItems = self._scanDatabase()
        for i in range(len(periodicFrequentItems)):
            itemX = periodicFrequentItems[i]
            if itemX is None:
                continue
            tidSetX = self._tidList[itemX]
            itemSetX = [itemX]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(periodicFrequentItems)):
                itemJ = periodicFrequentItems[j]
                if itemJ is None:
                    continue
                tidSetJ = self._tidList[itemJ]
                y1 = list(set(tidSetX).intersection(tidSetJ))
                if len(y1) < self._minSup:
                    continue
                if len(tidSetX) == len(tidSetJ) and len(y1) is len(tidSetX):
                    periodicFrequentItems.insert(j, None)
                    itemSetX.append(itemJ)
                elif len(tidSetX) < len(tidSetJ) and len(y1) is len(tidSetX):
                    itemSetX.append(itemJ)
                elif len(tidSetX) > len(tidSetJ) and len(y1) is len(tidSetJ):
                    periodicFrequentItems.insert(j, None)
                    itemSets.append(itemJ)
                    tidSets.append(y1)
                else:

                    itemSets.append(itemJ)
                    tidSets.append(y1)
            if len(itemSets) > 0:
                self._processEquivalenceClass(itemSetX, itemSets, tidSets)
            self._save([], itemSetX, tidSetX)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Closed periodic frequent patterns were generated successfully using CPFPMiner algorithm ")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataFrame

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.replace(' ', '\t').strip() + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Closed Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())
        

if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = CPFPMiner(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = CPFPMiner(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        print("Total number of Closed Periodic-Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



#  Copyright (C)  2021 Rage Uday Kiran
#
#             This program is free software: you can redistribute it and/or modify
#
#             it under the terms of the GNU General Public License as published by
#
#             the Free Software Foundation, either version 3 of the License, or
#
#             (at your option) any later version.
#
#             This program is distributed in the hope that it will be useful,
#
#             but WITHOUT ANY WARRANTY; without even the implied warranty of
#
#             MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#
#             GNU General Public License for more details.
#
#             You should have received a copy of the GNU General Public License
#
#             along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


import abstract as _hus
import pandas as pd
from functools import reduce
from operator import and_
from deprecated import deprecated

_minSup = str()
_hus._sys.setrecursionlimit(20000)


class _Node:
    """
    A class used to represent a node in the SHU tree

    :Attributes:

        itemName : str
            name of the item

        utility : int
            utility of the item

        children : dict
            dictionary of children of the node

        next : _Node
            pointer to the next node with same item in the tree

        batchIndex : int
            index of the batch with respect to window to which the node belongs

        utility : list
            list of utilities of the node with respect to each batch in the window

        parent : _Node
            pointer to the parent of the node

    :Methods:

        addUtility(utility, batchIndex)
            Adds utility to the node at specified batch index

        removeUtility(utility)
            Removes utility from the node

        shiftUtility()
            Shifts the utility values of the node to the left useful for removing the oldest batch information

        shiftTail()
            Shifts the tail values of the node to the left useful for removing the oldest batch information
    """

    def __init__(self, itemName, utility, batchSize, batchIndex):
        self.itemName = itemName
        self.utility = [0 for _ in range(batchSize)]
        self.children = dict()
        self.next = None
        self.batchIndex = batchIndex
        self.utility[batchIndex] = utility
        self.parent = None
        self.tail = None

    def addUtility(self, utility, batchIndex):
        """
        Adds utility to the node

        :param utility : Net utility value to be added

        :type utility : int

        :param batchIndex : index of the batch with respect to window to which the node belongs

        :type batchIndex : int
        """

        self.utility[batchIndex] += utility

    def removeUtility(self, utility):
        """
        Removes utility from the node

        :param utility: Net utility value to be removed

        :type utility : int
        """

        self.utility -= utility

    def shiftUtility(self):
        """
        Shifts the utility values of the node to the left
        """
        self.utility.pop(0)
        self.utility.append(0)

    def shiftTail(self):
        """
        Shifts the tail pointer of the node to the left
        """

        if(self.tail is not None):
            self.tail.pop(0)
            self.tail.append(False)


class _HeaderTable:
    """
    A class used to represent the header table of the SHU tree

    :Attributes:

        table : dict
            dictionary of items as keys and list of utility and pointer to the node in the tree as values
            representing the header table

        orderedItems : list
            List of items in the header table in lexicographical order

    :Methods:

        updateUtility(item, utility, node)
            Updates the utility of the item with node pointer in the header table

        addUtility(item, utility)
            Adds utility to the item in the header table

        removeUtility(item, utility)
            Removes utility from the item in the header table

        itemOrdering()
            Orders the items in the header table in lexicographical order
    """

    def __init__(self):
        self.table = dict()
        self.orderedItems = list()

    def updateUtility(self, item, utility, node):
        """
        Updates the utility of the item in the header table

        :param item: name of the item to which utility is to be updated

        :type item: str

        :param utility: Net utility of the item to be updated

        :type utility: int

        :param node: pointer to the node in the tree to which the item belongs

        :type node: _Node
        """

        if item in self.table:
            self.table[item][0] += utility
            tempNode = self.table[item][1]

            while tempNode.next is not None:
                tempNode = tempNode.next
            
            tempNode.next = node

        else:
            self.table[item] = [utility, node]
    
        self.itemOrdering()

    def addUtility(self, item, utility):
        """
        Adds utility to the item in the header table

        :param item: name of the item to which utility needs to be added

        :type item:str

        :param utility: Net utility of the item to be added

        :type utility: int
        """
        self.table[item][0] += utility
    

    def removeUtility(self, item, utility):
        """
        Removes utility from the item in the header table

        :param item: name of the item to which utility needs to be removed

        :type item: str

        :param: utility: Net utility of the item to be removed

        :type utility: int
        """
        self.table[item][0] -= utility

        if self.table[item][0] == 0:
            del self.table[item]

        self.itemOrdering()

    def itemOrdering(self):
        """
        Orders the items in the header table in lexicographical order
        """        
        self.orderedItems = list(sorted(self.table.keys()))

class _SHUTree:

    """
    A class used to represent the SHU tree

    :Attributes:

        root : _Node
            root node of the tree

        headerTable : _HeaderTable
            header table of the tree

        batchSize : int
            size of the batch

        windowSize : int
            size of the window

        batchIndex : int
            index of the current batch with respect to window

        windowUtility : int
            utility of the current window

        localTree : bool
            flag to indicate whether the tree is local useful for prefix/conditional tree generation

    :Methods:

        addTransaction(transaction, utility)
            Adds transaction to the tree

        tailUtilities(root)
            Returns the tail utilities of the tree

        removeBatch()
            Removes the oldest batch from the tree

        removeBatchUtility()
            Removes the utility of the oldest batch from each subtree

    """

    def __init__(self, batchSize, windowSize, localTree = False):
        self.root = _Node(None, 0, batchSize, 0)
        self.headerTable = _HeaderTable()
        self.batchSize = batchSize
        self.windowSize = windowSize
        self.batchIndex = 0
        self.windowUtility = 0
        self.localTree = localTree

    def addTransaction(self, transaction, utility, itemUtility = None):
        """
        Adds transaction to the tree

        :param transaction: list of items in the transaction

        :type transaction: list

        :param utility: Net utility of the transaction

        :type utility: int
        """
        # print("Transaction", transaction, itemUtility, self.localTree)
        transaction.sort(key = lambda x: x[0])
        currentNode = self.root
        self.windowUtility += utility

        curUtility = 0
        for iter in range(len(transaction)):
            
            item = transaction[iter]

            if(self.localTree is False):
                curUtility += itemUtility[iter]
            else:
                curUtility = itemUtility[iter]

            
            if item in currentNode.children:
                currentNode = currentNode.children[item]
                currentNode.addUtility(curUtility, self.batchIndex)

                if(self.localTree is False):
                    self.headerTable.addUtility(item, curUtility)

                else:
                    self.headerTable.addUtility(item, utility)

            else:
                newNode = _Node(item, curUtility, self.batchSize, self.batchIndex)
                currentNode.children[item] = newNode
                newNode.parent = currentNode

                if(self.localTree is False):
                    self.headerTable.updateUtility(item, curUtility, newNode)

                else:
                    self.headerTable.updateUtility(item, utility, newNode)
                
                currentNode = newNode

        currentNode.tail = [False for _ in range(self.batchSize)]
        currentNode.tail[self.batchIndex] = True

    def tailUtilities(self, root):
        """
        Calculates the utility of the tail of the node

        :param root: pointer to the node in the tree

        :type root: _Node

        :return: utility of the tail of the node

        :rtype: int
        """

        if(root is None):
            return 0

        if(root.tail is not None):
            return root.utility[0]

        netUtility = 0
        for item in root.children:
            netUtility += self.tailUtilities(root.children[item])

        return netUtility

    def removeBatch(self):

        """
        Removes the oldest batch from the tree
        """

        currentNode = self.root

        curChilds = list(currentNode.children.keys())

        for child in curChilds:
            self.windowUtility -= self.tailUtilities(currentNode.children[child])
            self.removeBatchUtility(currentNode.children[child])

            if(sum(currentNode.children[child].utility) == 0):
                del currentNode.children[child]


    def removeBatchUtility(self, tempNode = None):
        
        """
        Removes the utility of the oldest batch from each subtree

        :param tempNode: pointer to the node in the tree

        :type tempNode: _Node
        """
        if tempNode is None:
            return

        for item in tempNode.children:
            self.removeBatchUtility(tempNode.children[item])

        curBatchUtility = tempNode.utility[0]
        tempNode.shiftUtility()
        tempNode.shiftTail()

        if(sum(tempNode.utility) == 0):
            if(tempNode.itemName in self.headerTable.table):
                curNode = self.headerTable.table[tempNode.itemName][1]

                if(curNode == tempNode):
                    self.headerTable.table[tempNode.itemName][1] = tempNode.next

                else:
                    while(curNode != None and curNode.next != tempNode):
                        curNode = curNode.next

                    if(curNode != None):
                        curNode.next = tempNode.next

                del tempNode.tail
        
        self.headerTable.removeUtility(tempNode.itemName, curBatchUtility)

        curChilds = list(tempNode.children.keys())
        for child in curChilds:
            if(sum(tempNode.children[child].utility) == 0):
                del tempNode.children[child]

    
class SHUGrowth(_hus._highUtilityPatternStreamMining):
    """
    :Description:   High-utility pattern mining over data stream is one of the challenging problems in data stream mining.
                    SHUGrowth is an algorithm that discovers high-utility patterns from data streams without rebuilding the tree.
                    It stores the database of the current window in form of SHUTree and adjusts the tree based on upcoming
                    transactions removing the oldest batch. It is an optimized varaint of HUPMS algorithm.

    :References:   Chowdhury Farhan Ahmed and Syed Khairuzzaman Tanbeer and Byeong-Soo Jeong and Ho-Jin Choi : High utility
                   pattern mining over data streams with sliding window technique. Expert Systems with Applications Vol 57,
                   214 - 231, 2016. https://doi.org/10.1016/j.eswa.2016.03.001

    :param  iFile: str :
                   Name of the Input file to mine complete set of High Utility patterns in Streams
    :param  oFile: str :
                   Name of the output file to store complete set of High Utility patterns in Streams
    :param minUtil: int :
                   Minimum utility threshold
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Attributes:

        __startTime : float
            start time of the mining process

        __endTime : float
            end time of the mining process

        _minUtil : str
            minimum utility threshold

        __finalPatterns : dict
            dictionary of the final patterns generated in each window

        _iFile : str
            input file name

        _oFile : str
            output file name

        _sep : str
            seperator of the input file

        __memoryUSS : float
            memory usage of the algorithm in user space

        __memoryRSS : float
            memory usage of the algorithm in resident set size

        __transactions : list
            list of transactions in the database

        __utilities : list
            list of utilities of each item of a transaction in the database

            __utilitySum : list
                list of utility sum of each transaction in the database

            __tree : _SHUTree
                SHU tree of the current window

            __windowSize : int
                The size of the sliding window. It specifies the number of panes to be considered for mining patterns.

            __paneSize : int
                The size of the pane. It specifies the number of transactions to be considered for mining in each pane.

    :Methods:

        _createItemsets()
            Storing the complete transactions of the database/input file in a transaction variable with their utilities.

        minPathUtil(nodeIndex, stack)
            Calculates the minimum utility of the path from the root to the ends of the tree

        createPrefixBranch(root)
            Creates the prefix branch of the current SHU-Tree for construction of prefix tree

        fixUtility(root)
            Fixes the utility of the nodes in the tree by merging the utility values

        createConditionalTree(root, transactions, minUtil)
            Creates the conditional tree for the given prefix tree

        contains(superset, subset)
            Checks if the superset contains the subset

        treeGenerations(root, netUtil, candidatePattern, curItem)
            Generates the tree of the high utility patterns

        mine()
            Starts the mining process

        printTree(root, level)
            Prints the SHU-tree in a readable format

        getMemoryRSS()
            Returns the memory usage of the algorithm in resident set size

        getMemoryUSS()
            Returns the memory usage of the algorithm in user space

        getPatterns()
            Returns the final patterns generated in each window

        getPatternsAsDataFrame()
            Returns the final patterns generated in each window as a pandas dataframe

        getRuntime()
            Returns the runtime of the algorithm

        printResults()
            Prints the stats of the mining process

        save()
            Saves the patterns generated from each window in a file


    ** Executing the code on terminal:**
    ---------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 SHUGrowth.py <inputFile> <outputFile> <minUtil> <windowSize> <paneSize> <separator>

      Example Usage:

      (.venv) $ python3 SHUGrowth.py retail.txt output.txt 107 100 1000 ','

    .. note:: Here minimum utility is 107, Window size is 100 and pane size is 1000. The separator is comma for the input file


    **Credits:**
    --------------
    The code is written by Vipul Chhabra under the supervision of Prof. Rage Uday Kiran.
    
    """

    __startTime = float()
    __endTime = float()
    _minUtil = str()
    __finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    __memoryUSS = float()
    __memoryRSS = float()
    __transactions = []
    __utilities = []
    __utilitySum = []
    __tree = None
    __windowSize = 0
    __paneSize = 0

    def __init__(self, iFile, oFile, minUtil, windowSize, paneSize, sep = ","):
        super().__init__(iFile, minUtil, windowSize, paneSize, sep)
        self._oFile = oFile

    def _createItemsets(self):
        """
        Storing the complete transactions of the database/input file in a transaction variable
        """
        self._transactions, self._utilities, self._utilitySum = [], [], []
        if isinstance(self._iFile, _hus._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._transactions = self._iFile['Transactions'].tolist()
            if 'Utilities' in i:
                self._utilities = self._iFile['Utilities'].tolist()
            if 'UtilitySum' in i:
                self._utilitySum = self._iFile['UtilitySum'].tolist()
        if isinstance(self._iFile, str):
            if _hus._validators.url(self._iFile):
                data = _hus._urlopen(self._iFile)
                for line in data:
                    line = line.decode("utf-8")
                    line = line.split("\n")[0]
                    parts = line.split(":")
                    items = parts[0].split(self._sep)
                    self.__transactions.append([x for x in items if x])
                    utilities = parts[2].split(self._sep)
                    utilities = [float(x) for x in utilities]
                    self.__utilities.append(utilities)
                    self.__utilitySum.append(int(parts[1]))
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.split("\n")[0]
                            parts = line.split(":")
                            items = parts[0].split(self._sep)
                            self._transactions.append([x for x in items if x])
                            utilities = parts[2].split(self._sep)
                            utilities = [float(x) for x in utilities]
                            self._utilities.append(utilities)
                            self._utilitySum.append(float(parts[1]))
                except IOError:
                    print("File Not Found")
                    quit()

    def minPathUtil(self, nodeIndex, stack):
        """
        Calculates the minimum utility of the path from the root to the ends of the tree

        :param nodeIndex: index of the node in the stack

        :type nodeIndex: int

        :param stack: list of the nodes in the prefix branch

        :type stack: list
        """

        minUtil = 0
        activeBatch = [i for i, e in enumerate(stack[0].utility) if e != 0]

        for batch in activeBatch:
            if(stack[nodeIndex + 1].tail == None or stack[nodeIndex + 1].tail[batch] == False):
                minUtil += (stack[nodeIndex].utility[batch] - stack[nodeIndex + 1].utility[batch])
            
        return minUtil
        

    def createPrefixBranch(self, root):
        """
        Creates the prefix branch of the node

        :param root: pointer to the root node of the sub-tree

        :type root: _Node

        :return: list of the nodes in prefix branch

        :rtype: list

        :return: utility of the prefix branch

        :rtype: int
        """
        stack = []

        while(root is not None):
            stack.append(root)
            root = root.parent

        chosenItemset = stack[0]
        lastUtil = sum(chosenItemset.utility)
        curBacthes = [i for i, e in enumerate(chosenItemset.utility) if e != 0]

        otherUtilites = []

        for i in range(1, len(stack)-1):
            curItemset = stack[i]
            epu = lastUtil

            for j in range(i-1, 0, -1):
                epu -= self.minPathUtil(j, stack)
            
            otherUtilites.append(epu)

        return stack, lastUtil, otherUtilites

    def fixUtility(self, root):
        """
        Fixes the utility of the nodes in the tree by merging the utility values

        :param root: pointer to the root node of the sub-tree

        :type root: _Node
        """
        
        if(root is None):
            return
        
        if(len(root.utility) > 1):
            root.utility = [sum(root.utility)]

        for child in root.children:
            self.fixUtility(root.children[child])

    def createConditionalTree(self, root, transactions, minUtil):
        """
        Creates the conditional tree for the given prefix tree

        :param root: pointer to the root node of the prefix tree

        :type root: _Node

        :param transactions: list of transactions in prefix tree

        :type transactions: list

        :param minUtil: minimum utility threshold

        :type minUtil: int

        :return: tempTree: conditional tree for the given prefix tree

        :rtype: _SHUTree
        """
        
        for transaction in transactions:
            for item in transaction["transaction"]:
                if(root.headerTable.table[item][0] < minUtil):
                    itemIndex = transaction["transaction"].index(item)
                    transaction["transaction"].remove(item)
                    del transaction["itemwiseUtility"][itemIndex]

        tempTree = _SHUTree(1, 1, True)
        for transaction in transactions:
            if(len(transaction["transaction"]) != 0):
                tempTree.addTransaction(transaction["transaction"], transaction["utility"], transaction["itemwiseUtility"])


        return tempTree
    
    def contains(self, superset, subset):
        """
        Checks if the superset contains the subset

        :param superset: list of items in the superset

        :type superset: list

        :param subset: list of items in the subset

        :type subset: list

        :return: True if superset contains subset, False otherwise

        :rtype: bool
        """
     
        return reduce(and_, [i in superset for i in subset])

    def treeGenerations(self, root, netUtil, candidatePattern, curItem = []):
        """
        Generates the tree of the high utility patterns

        :param root: pointer to the root of the tree

        :type root: _Node

        :param netUtil: Net utility of the transaction

        :type netUtil: int

        :param candidatePattern: Candidate patterns generated with utility

        :type candidatePattern: list

        :param curItem: List of items in the current itemsets

        :type curItem: list
        """

        if(root is None):
            return


        for item in reversed(root.headerTable.orderedItems):
            if(root.headerTable.table[item][0] >= netUtil):
                prefixBranches = []

                tempNode = root.headerTable.table[item][1]

                while tempNode is not None:
                    curPrefixBranch, lastUtil, otherUtilites = self.createPrefixBranch(tempNode)
                    prefixBranches.append([curPrefixBranch, lastUtil, otherUtilites])
                    tempNode = tempNode.next
                
                prefixTree = _SHUTree(1, 1, True)
                completeTransactions = []

                for branch in prefixBranches:
                    transaction = []
                    utilities = []
                    for idx in range(len(branch[2])-1, -1, -1):
                        transaction.append(branch[0][(idx + 1)].itemName)
                        utilities.append(branch[2][idx])

                    prefixTree.addTransaction(transaction, branch[1], utilities)

                    completeTransactions.append({"transaction" : transaction, "utility" : branch[1], "itemwiseUtility" : utilities})

                conditionalTree = self.createConditionalTree(prefixTree, completeTransactions, netUtil)

                newItemset = curItem.copy()
                newItemset.append(item)

                if(len(newItemset) not in candidatePattern):
                    candidatePattern[len(newItemset)] = [newItemset]

                else:
                    candidatePattern[len(newItemset)].append(newItemset)

                if(len(conditionalTree.headerTable.table) != 0):
                    self.treeGenerations(conditionalTree, netUtil, candidatePattern, newItemset)

    @deprecated("It is recommended to use 'mine()' instead of 'startMine()' for mining process. Starting from January 2025, 'startMine()' will be completely terminated.")
    def startMine(self):
        """
        This function will start the mining process
        """
        self.mine()

    def mine(self):
        """
        This function will start the mining process
        """
        global _minUtil
        self.__startTime = _hus._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minUtil is None:
            raise Exception("Please enter the Minimum Support")
        if self._windowSize is None:
            raise Exception("Please enter the Window Size")
        if self._paneSize is None:
            raise Exception("Please enter the Pane Size")
        self.__windowSize = int(self._windowSize)
        self.__paneSize = int(self._paneSize)

        self._createItemsets()
        self._minUtil = float(self._minUtil)
        self.__tree = _SHUTree(self.__windowSize, self.__paneSize)

        transactionwiseUtility = []

        for i in range(len(self._transactions)):
            curTrans = {}
            for j in range(len(self._transactions[i])):
                curTrans[self._transactions[i][j]] = self._utilities[i][j]
            transactionwiseUtility.append(curTrans)

        for i in range(0, self.__windowSize):
            self.__tree.batchIndex = i
            for j in range(0, self.__paneSize):
                self.__tree.addTransaction(self._transactions[i * self.__paneSize + j],
                                           self._utilitySum[i * self.__paneSize + j],
                                           self._utilities[i * self.__paneSize + j])

        startIndex = 0
        endIndex = self.__windowSize * self.__paneSize

        while (endIndex <= len(self._transactions)):

            filteredItemsets = {}

            self.treeGenerations(self.__tree, self._minUtil, filteredItemsets)

            results = []

            for itemSetLen in filteredItemsets:
                for itemSet in filteredItemsets[itemSetLen]:
                    itemSetUtility = 0
                    for transId in range(startIndex, endIndex):
                        if (self.contains(list(transactionwiseUtility[transId].keys()), itemSet)):
                            for item in itemSet:
                                itemSetUtility += transactionwiseUtility[transId][item]

                    if (itemSetUtility >= self._minUtil):
                        results.append([itemSet, itemSetUtility])

            self.__finalPatterns[(startIndex, endIndex)] = results

            if (endIndex >= len(self._transactions)):
                break

            self.__tree.removeBatch()

            for i in range(0, self.__paneSize):
                self.__tree.addTransaction(self._transactions[endIndex + i], self._utilitySum[endIndex + i],
                                           self._utilities[endIndex + i])

            startIndex += self.__paneSize
            endIndex += self.__paneSize

        self.__endTime = _hus._time.time()
        self.__memoryUSS = float()
        self.__memoryRSS = float()
        process = _hus._psutil.Process(_hus._os.getpid())
        self.__memoryUSS = process.memory_full_info().uss
        self.__memoryRSS = process.memory_info().rss

    def printTree(self, root, level = 0):
        """
        Prints the tree in a readable format.

        :param root: CPSTreeNode object for the root of the tree

        :type root: CPSTreeNode

        :param level: Current level of the root node

        :type level: int
        """

        print('  ' * level, level, root.itemName, root.utility, root.parent.itemName if root.parent else None )
        
        if(root.tail is not None):
            print('  ' * (level + 1), level + 1, root.tail)

        for child in root.children.values():
            self.printTree(child, level + 1)

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self.__memoryRSS

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """
        return self.__memoryUSS

    def getPatterns(self):
        """
        Returns the frequent patterns generated by the mining process over the complete datastream.

        :return: returning frequent patterns generated by the mining process
        :rtype: dict
        """
        return self.__finalPatterns

    def getPatternsAsDataFrame(self):
        """
        Stores the final patterns generated by the mining process in a dataframe.

        :return: returning dataframe containing the final patterns generated by the mining process
        :rtype: pandas.DataFrame
        """

        dataframe = {}
        data = []
        for x, y in self.__finalPatterns.items():
            for pattern in y:
                patternString = ' '.join(pattern[0])
                data.append([x[0], x[1], patternString, pattern[1]])
        dataframe = pd.DataFrame(data, columns=['Window Start Index', 'Window End Index', 'Pattern', 'Utility'])
        return dataframe

    def getRuntime(self):
        """
        Total amount of time taken by the mining process will be retrieved from this function

        :return: returning time taken by the mining process
        :rtype: float
        """

        return self.__endTime - self.__startTime

    def printResults(self):
        """
        Prints the stats of the mining process
        """
        print("Total number of Windows Processed:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


    def save(self):
        """
        Complete set of frequent patterns will be loaded in to a output file
        """
        print("Output file name", self._oFile)
        writer = open(self._oFile, 'w+')
        for x, y in self.__finalPatterns.items():
            writer.write("Window Start Index : %s , End Index : %s \n" % (x[0], x[1]))
            for pattern in y:
                patternString = '\t'.join(pattern[0])
                patternString += '\t' + ":" + '\t' + str(pattern[1])
                writer.write("%s \n" % patternString)


if __name__ == "__main__":
    _ap = str()
    if len(_hus._sys.argv) == 6 or len(_hus._sys.argv) == 7:
        if len(_hus._sys.argv) == 7:
            _ap = SHUGrowth(_hus._sys.argv[1], _hus._sys.argv[2], _hus._sys.argv[3], _hus._sys.argv[4], _hus._sys.argv[5], _hus._sys.argv[6])
        if len(_hus._sys.argv) == 6:
            _ap = SHUGrowth(_hus._sys.argv[1], _hus._sys.argv[2], _hus._sys.argv[3], _hus._sys.argv[4], _hus._sys.argv[5])
        _ap.startMine()
        _ap.mine()
        print("Total number of Windows Processes:", len( _ap.getPatterns()))
        _ap.getPatternsAsDataFrame().to_csv("result.csv", index = False, sep='\t')
        _ap.save()
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")


# PFECLAT is the fundamental approach to mine the periodic-frequent patterns.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.periodicFrequentPattern.basic import PFECLAT as alg
#
#             iFile = 'sampleDB.txt'
#
#             minSup = 10  # can also be specified between 0 and 1
#
#             maxPer = 20 # can also be specified between 0 and 1
#
#             obj = alg.PFECLAT(iFile, minSup, maxPer)
#
#             obj.mine()
#
#             periodicFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))
#
#             obj.save("periodicFrequentPatterns")
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

import pandas as pd
from deprecated import deprecated
import numpy as np

from PAMI.periodicFrequentPattern.basic import abstract as _ab


class PFECLAT(_ab._periodicFrequentPatterns):
    """
    **About this algorithm**

    :**Description**:   PFECLAT is the fundamental approach to mine the periodic-frequent patterns.

    :**Reference**:   P. Ravikumar, P.Likhitha, R. Uday kiran, Y. Watanobe, and Koji Zettsu, "Towards efficient discovery of
                      periodic-frequent patterns in columnar temporal databases", 2021 IEA/AIE.

    :**Parameters**:    - **iFile** (*str or URL or dataFrame*) -- *Name of the Input file to mine complete set of frequent patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of frequent patterns.*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.*
                        - **maxPer** (*int or float or str*) -- *The user can specify maxPer either in count or proportion of database size. It controls the maximum number of transactions in which any two items within a pattern can reappear.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **Database** (*list*) -- *To store the transactions of a database in list.*
                        - **mapSupport** (*Dictionary*) -- *To maintain the information of item and their frequency.*
                        - **lno** (*int*) -- *It represents the total no of transactions*
                        - **tree** (*class*) -- *it represents the Tree class.*
                        - **itemSetCount** (*int*) -- *it represents the total no of patterns.*
                        - **tidList** (*dict*) -- *stores the timestamps of an item.*
                        - **hashing** (*dict*) -- *stores the patterns with their support to check for the closed property.*

    :**Methods**:       - **startMine()** -- *Mining process will start from here.*
                        - **getPatterns()** -- *Complete set of patterns will be retrieved with this function.*
                        - **save(oFile)** -- *Complete set of periodic-frequent patterns will be loaded in to a output file.*
                        - **getPatternsAsDataFrame()** -- *Complete set of periodic-frequent patterns will be loaded in to a dataframe.*
                        - **getMemoryUSS()** -- *Total amount of USS memory consumed by the mining process will be retrieved from this function.*
                        - **getMemoryRSS()** -- *Total amount of RSS memory consumed by the mining process will be retrieved from this function.*
                        - **getRuntime()** -- *Total amount of runtime taken by the mining process will be retrieved from this function.*
                        - **creatingOneItemSets()** -- *Scan the database and store the items with their timestamps which are periodic frequent.*
                        - **getPeriodAndSupport()** -- *Calculates the support and period for a list of timestamps.*
                        - **Generation()** -- *Used to implement prefix class equivalence method to generate the periodic patterns recursively*

    **Execution methods**

    **Terminal command**

    .. code-block:: console

       Format:

       (.venv) $ python3 PFECLAT.py <inputFile> <outputFile> <minSup> <maxPer>

       Example usage:

       (.venv) $ python3 PFECLAT.py sampleDB.txt patterns.txt 10.0 20.0

    .. note:: minSup will be considered in percentage of database transactions


    **Calling from a python program**

    .. code-block:: python

            from PAMI.periodicFrequentPattern.basic import PFECLAT as alg

            iFile = 'sampleDB.txt'

            minSup = 10  # can also be specified between 0 and 1

            maxPer = 20 # can also be specified between 0 and 1

            obj = alg.PFECLAT(iFile, minSup, maxPer)

            obj.mine()

            periodicFrequentPatterns = obj.getPatterns()

            print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))

            obj.save("periodicFrequentPatterns")

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**

    The complete program was written by P. Likhitha  and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """
    
    _iFile = " "
    _oFile = " "
    _sep = " "
    _dbSize = None
    _Database = None
    _minSup = str()
    _maxPer = str()
    _tidSet = set()
    _finalPatterns = {}
    _startTime = None
    _endTime = None
    _memoryUSS = float()
    _memoryRSS = float()

    def _convert(self, value) -> float:
        """
        To convert the given user specified value

        :param value: user specified value
        :type value: int or float or str
        :return: converted value
        :rtype: int or float
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._dbSize * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._dbSize * value)
            else:
                value = int(value)
        return value

    def _creatingItemSets(self) -> None:
        """

        Storing the complete transactions of the database/input file in a database variable

        :return: None
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                if data[i]:
                    tr = [str(ts[i])] + [x for x in data[i].split(self._sep)]
                    self._Database.append(tr)
                else:
                    self._Database.append([str(ts[i])])

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    @deprecated("It is recommended to use 'mine()' instead of 'startMine()' for mining process. Starting from January 2025, 'startMine()' will be completely terminated.")
    def startMine(self) -> None:
        self.mine()

    def _getMaxPer(self, arr, maxTS):
        arr = np.append(list(arr), [0, maxTS])
        arr = np.sort(arr)
        arr = np.diff(arr)

        return np.max(arr)

    def mine(self) -> None:
        """
        Mining process will start from this function
        :return: None
        """
        self._startTime = _ab._time.time()
        self._finalPatterns = {}
        frequentSets = self._creatingItemSets()

        items = {}
        maxTS = 0
        for line in self._Database:
            index = int(line[0])
            maxTS = max(maxTS, index)
            for item in line[1:]:
                if tuple([item]) not in items:
                    items[tuple([item])] = set()
                items[tuple([item])].add(index)

        self._dbSize = maxTS

        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        minSup = self._minSup
        maxPer = self._maxPer


        items = {k: v for k, v in items.items() if len(v) >= minSup}
        items = {k: v for k, v in sorted(items.items(), key = lambda x: len(x[1]), reverse = True)}

        keys = []
        for item in list(items.keys()):
            per = self._getMaxPer(items[item], maxTS)
            if per <= maxPer:
                keys.append(item)
                self._finalPatterns[item] = [len(items[item]), per, set(items[item])]

        while keys:
            newKeys = []
            for i in range(len(keys)):
                for j in range(i + 1, len(keys)):
                    if keys[i][:-1] == keys[j][:-1] and keys[i][-1] != keys[j][-1]:
                        # print(keys[i], keys[j])
                        newKey = tuple(keys[i] + (keys[j][-1],))
                        intersect = items[keys[i]].intersection(items[keys[j]])
                        per = self._getMaxPer(intersect, maxTS)
                        sup = len(intersect)
                        if sup >= minSup and per <= maxPer:
                            items[newKey] = intersect
                            newKeys.append(newKey)
                            self._finalPatterns[newKey] = [sup, per, set(intersect)]
                    else:
                        break
            keys = newKeys

        newPattern = {}
        for k, v in self._finalPatterns.items():
            newPattern["\t".join([str(x) for x in k])] = v

        self._finalPatterns = newPattern

        # self._generateEclat(frequentSets)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Periodic-Frequent patterns were generated successfully using PFECLAT algorithm ")

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self) -> _ab._pd.DataFrame:
        """
        Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataframe

    def save(self, outFile: str) -> None:
        """
        Complete set of periodic-frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: csv file
        :return: None
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y[0]) + ":" + str(y[1])
            #s1 = x.replace(' ', '\t') + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % s1)

    def getPatterns(self) -> dict:
        """
        Function to send the set of periodic-frequent patterns after completion of the mining process

        :return: returning periodic-frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self) -> None:
        """
        This function is used to print the results
        :return: None
        """
        print("Total number of Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())
                    

if __name__ == "__main__":



    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PFECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PFECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        _ap.mine()
        print("Total number of Periodic-Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")

    


# GTCP is graph transactional coverage pattern mining algorithm which esentially computes the coverage patterns for graph transactional data with applications in Drug discovery 
# **Importing this algorithm into a python program**
# --------------------------------------------------------

#       from PAMI.GraphTransactionalCoveragePattern.basic.GTCP import GTCP
#
#       fname="Toydata.txt"
#
#       obj=GTCP(fname,0.1,0.1,0.3,0.2)
#
#       obj.mine()  
#
#       obj.getPatterns()
#
#       memUSS = obj.getMemoryUSS()
#
#       obj.save(oFile)
#
#       print("Total Memory in USS:", memUSS)
#
#       memRSS = obj.getMemoryRSS()
#
#       print("Total Memory in RSS", memRSS)
#
#       run = obj.getRuntime()
#
#       print("Total ExecutionTime in seconds:", run)



from bitarray import bitarray
from PAMI.subgraphMining.basic import gspan as gsp
from PAMI.extras.stats import graphDatabase as gdb
from PAMI.GraphTransactionalCoveragePattern.basic import abstract as _ab

class GTCP:
    def __init__(self,iFile,minsup,minGTC,minGTPC,maxOR=0.2):
        """
            iFile : input file
            minsup : Minimum support 
            minGTC : Minimum Graph transaction coverage
            minGTPC : Minimum graph pattern coverage 
            maxOR : Maximum overlap ratio
            Sf : subgraphsBygraphID
            Df: Flat transactional Dataset
        """

        self.Df=[]
        self.Sf=[]
        self.L={}
        self.iFile=iFile
        self.maxOR=maxOR
        self.minGTC=minGTC 
        self.minGTPC=minGTPC
        gdb_obj=gdb.graphDatabase(self.iFile)
        self.numGraphs=len(gdb_obj.graphs)
        gsp_obj = gsp.GSpan(self.iFile, minsup, outputSingleVertices=False, maxNumberOfEdges=float('inf'), outputGraphIds=True)
        gsp_obj.mine()
        self.Sf=gsp_obj.getSubgraphGraphMapping()
        self.GetFIDBasedFlatTransactions()
        # print("Subgraph mining completed")

    def mine(self):
        """
            Mine the coverage patterns
        """
        self._startTime = _ab._time.time()
        self.Cmine()
    

    def getPatterns(self):
        """
            Get all the coverage patterns
        """
        return self.L

    def Coverage(self,g):
        """
            Get coverage of a graph
            param
                g : Graph id
        """
        return self.Df[g].count()/len(self.Sf)


    def patternCoverage(self,pattern):
        """
            Get patternCoverage of a pattern
            param
                pattern: pattern for which pattern coverage needs to be computed
        """
        Csets=list(map(lambda x:self.Df[x] ,pattern))
        newset=bitarray(len(self.Sf))
        for coverage in Csets:
            newset=newset | coverage
        return newset

    def OverlapRatio(self,pattern):
        """
            Get Overlap ratio of a pattern
            param
                pattern: pattern for which overlap ratio needs to be computed
        """

        lastitem=pattern[-1]
        lastbut=pattern[:-1]
        lastbutcoverage=self.patternCoverage(lastbut)
        lastcoverage=self.Df[lastitem]
        
        intersection=lastcoverage & lastbutcoverage
        cs= (lastcoverage | lastbutcoverage).count()/len(self.Sf)
        return (intersection.count()/self.Df[lastitem].count(),cs)


    def GetFIDBasedFlatTransactions(self):
        """
            Convert into FID based transactions
        """
        self.Df={}
        for i in range(self.numGraphs):
            bitset = bitarray(len(self.Sf))
            bitset.setall(0)
            self.Df.update({i:bitset})

        for fragment in self.Sf:
            for GID in fragment["GIDs"]:
                self.Df[GID][fragment["FID"]]=1

    def getallFreq1(self):
        """
            Get all the Patterns of size 1

        """
        freq=[(graph,self.Coverage(graph)) for graph in self.Df if self.Coverage(graph)>=self.minGTC]

        sorted_list = sorted(freq, key=lambda x: x[1], reverse=True)
        final_list=list(map(lambda x: [x[0]],sorted_list))
        return final_list

    def join(self,l1,l2):
        """
            Join two patterns
            Param 
                l1: Pattern 1
                l2: Pattern 2
        """
        patterns=[]
        Nol=[]
        temp_l=[]
        newpattern=[]
        for i in range(len(l1)):
            for j in range(i+1,len(l2)):
                if(l1[i][:-1]==l2[j][:-1]):
                    if(self.Coverage(l1[i][-1])>=self.Coverage(l2[j][-1])):
                        newpattern= l1[i]+[l2[j][-1]]
                    else:
                        newpattern=l2[j]+[l1[i][-1]]
                    
                    ov,cs=self.OverlapRatio(newpattern)
                    
                    if(ov<=self.maxOR):
                        if(cs>=self.minGTPC):
                            self.L.append((newpattern,cs))
                        else:
                            self.Nol.append(newpattern)
                else:
                    break
        # return (Nol,temp_l)    
    

    def writePatterns(self):
        outf=self.iFile.split(".")[0]+"_results.txt"
        f=open(outf,"w+")
        for i in self.L:
            f.write(str(i[0])+", Coverage : "+str(i[1]))
            f.write("\n")
        
        f.close()
    
    def Cmine(self):
        """
        Cmine Algorithm for mining coverage patterns
        """

        self.Nol_1=self.getallFreq1()
        l=1
        self.L=[]
        self.Nol_1_temp=[]
        for g in self.Nol_1:
            if(self.Coverage(g[0])>=self.minGTPC):
                self.L.append((g,self.Coverage(g[0])))
            else:
                self.Nol_1_temp.append(g)
        l+=1
        self.Nol_1=self.Nol_1_temp
        self.Nol=[]


        while(len(self.Nol_1)>0):
            self.Nol=[]
            # print(len(self.Nol_1))
            self.join(self.Nol_1,self.Nol_1)
            self.Nol_1=self.Nol
            l+=1

        process = _ab._psutil.Process(_ab._os.getpid())
        self._endTime = _ab._time.time()
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
    
    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime-self._startTime


# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.periodicFrequentPattern.basic import parallelPFPGrowth as alg
#
#             obj = alg.parallelPFPGrowth(iFile, minSup, maxPer, noWorkers)
#
#             obj.startMine()
#
#             periodicFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))
#
#             obj.savePatterns(oFile)
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#



__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.periodicFrequentPattern.pyspark import abstract as _ab
from pyspark import SparkContext, SparkConf

from PAMI.periodicFrequentPattern.basic import abstract as _ab
import pandas as pd
from deprecated import deprecated

_maxPer = float()
_minSup = float()
_lno = int()


class Node(object):
    """
    A class used to represent the node of frequentPatternTree

    :Attributes:

        item : int or None
            Storing item of a node
        timeStamps : list
            To maintain the timestamps of a database at the end of the branch
        parent : node
            To maintain the parent of every node
        children : list
            To maintain the children of a node
        count : int
            To maintain the count of every node

    :Methods:

        addChild(itemName)
            Storing the children to their respective parent nodes
        toString()
            To print the node
    """

    def __init__(self, item, count, children):
        """ Initializing the Node class

        :param item: item of a node
        :param count: count of a node
        :param children: children of a node

        """
        self.item = item
        self.count = count
        self.children = children  # dictionary of children
        self.parent = None
        self.tids = set()

    def __repr__(self):
        return self.toString(0)

    def toString(self, level=0):
        """ To print the node

        :param level: level of a node
        """
        if self.item == None:
            s = "Root("
        else:
            s = "(item=" + str(self.item)
            s += ", count=" + str(self.count)
            for i in self.tids:
                s += " " + str(i)
        tabs = "\t".join(['' for i in range(0, level + 2)])
        for v in self.children.values():
            s += tabs + "\n"
            s += tabs + v.toString(level=level + 1)
        s += ")"
        return s

    def addChild(self, node):
        """ To add the children to a node

        :param node: children of a node
        """
        self.children[node.item] = node
        node.parent = self

    def _getTransactions(self):
        count = self.count
        tids = self.tids
        for child in self.children.values():
            for t in child._getTransactions():
                count -= t[2]
                t[0].insert(0, child.item)
                yield t
        if (count > 0):
            yield ([], tids, count)


class PFPTree(object):
    """

    A class used to represent the periodic frequent pattern tree

    :Attributes:

        root : node
            To maintain the root of the tree
        summaries : dict
            To maintain the summary of the tree

    :Methods:

        add(basket, tid, count)
            To add the basket to the tree
        getTransactions()
            To get the transactions of the tree
        merge(tree)
            To merge the tree
        project(itemId)
            To project the tree
        satisfyPer(tids, maxPer, numTrans)
            To satisfy the periodicity constraint
        extract(minCount, maxPer, numTrans, isResponsible = lambda x:True)
            To extract the periodic frequent patterns

    """

    def __init__(self):
        self.root = Node(None, 0, {})
        self.summaries = {}

    def __repr__(self):
        return repr(self.root)

    def add(self, basket, tid, count):
        """
        To add the basket to the tree

        :param basket: basket of a database
        :param tid: timestamp of a database
        :param count: count of a node
        """
        curr = self.root
        curr.count += count
        for i in tid:
            curr.tids.add(i)

        for i in range(0, len(basket)):
            item = basket[i]
            if item in self.summaries.keys():
                summary = self.summaries.get(item)
            else:
                summary = Summary(0, set())
                self.summaries[item] = summary
            summary.count += count

            if item in curr.children.keys():
                child = curr.children.get(item)
            else:
                child = Node(item, 0, {})
                curr.addChild(child)
            summary.nodes.add(child)
            child.count += count
            for j in tid:
                summary.tids.add(j)
                if (i == len(basket) - 1):
                    child.tids.add(j)
            curr = child
        return self

    def getTransactions(self):
        """
        To get the transactions of the tree

        :return: returning the transactions of the tree
        """
        return [x for x in self.root._getTransactions()]

    def merge(self, tree):
        """
        To merge the tree

        :param tree: tree of a database
        """
        for t in tree.getTransactions():
            self.add(t[0], t[1], t[2])
        return self

    def project(self, itemId):
        """
        To project the tree

        :param itemId: item of a node
        """
        newTree = PFPTree()
        summaryItem = self.summaries.get(itemId)
        if summaryItem:
            for element in summaryItem.nodes:
                t = []
                curr = element.parent
                while curr.parent:
                    t.insert(0, curr.item)
                    curr = curr.parent
                newTree.add(t, element.tids, element.count)
        return newTree

    def satisfyPer(self, tids, maxPer, numTrans):
        """
        To satisfy the periodicity constraint

        :param tids: timestamps of a database
        :param maxPer: maximum periodicity
        :param numTrans: number of transactions
        """

        tids = list(tids)
        tids.sort()
        if tids[0] > maxPer:
            return 0
        tids.append(numTrans)
        for i in range(1, len(tids)):
            if (tids[i] - tids[i - 1]) > maxPer:
                return 0
        return 1

    def extract(self, minCount, maxPer, numTrans, isResponsible=lambda x: True):
        """
        To extract the periodic frequent patterns

        :param minCount: minimum count of a node
        :param maxPer: maximum periodicity
        :param numTrans: number of transactions
        :param isResponsible: responsible node of a tree
        """
        for item in sorted(self.summaries, reverse=True):
            summary = self.summaries[item]
            if (isResponsible(item)):
                if (summary.count >= minCount and self.satisfyPer(summary.tids, maxPer, numTrans)):
                    yield ([item], summary.count)
                    for element in self.project(item).extract(minCount, maxPer, numTrans):
                        yield ([item] + element[0], element[1])
            for element in summary.nodes:
                parent = element.parent
                parent.tids |= element.tids


class Summary(object):
    """
    A class used to represent the summary of the tree

    :Attributes:

        count : int
            To maintain the count of a node
        nodes : list
            To maintain the nodes of a tree
        tids : set
            To maintain the timestamps of a database

    """

    def __init__(self, count, nodes):
        self.count = count
        self.nodes = nodes
        self.tids = set()


class parallelPFPGrowth(_ab._periodicFrequentPatterns):
    """
    :Description:   ParallelPFPGrowth is one of the fundamental distributed algorithm to discover periodic-frequent patterns in a transactional database. It is based PySpark framework.

    :Reference:   C. Saideep, R. Uday Kiran, Koji Zettsu, Cheng-Wei Wu, P. Krishna Reddy, Masashi Toyoda, Masaru Kitsuregawa: Parallel Mining of Partial Periodic Itemsets in Big Data. IEA/AIE 2020: 807-819

    :param  iFile: str :
                   Name of the Input file to mine complete set of periodic frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of periodic frequent pattern's
    :param  minSup: str:
                   Controls the minimum number of transactions in which every item must appear in a database.
    :param  maxPer: str:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        iFile : file
            Name of the Input file or path of the input file
        oFile : file
            Name of the output file or path of the output file
        minSup: int or float or str
            The user can specify minSup either in count or proportion of database size.
            If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
            Otherwise, it will be treated as float.
            Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
        maxPer: int or float or str
            The user can specify maxPer either in count or proportion of database size.
            If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
        numWorker: int
            The user can specify the number of worker machines to be employed for finding periodic-frequent patterns.
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            To represent the total no of transaction
        tree : class
            To represents the Tree class
        itemSetCount : int
            To represents the total no of patterns
        finalPatterns : dict
            To store the complete patterns

    :Methods:

        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of periodic-frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of periodic-frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets(fileName)
            Scans the dataset and stores in a list format
        PeriodicFrequentOneItem()
            Extracts the one-periodic-frequent patterns from database
        updateDatabases()
            Update the database by removing aperiodic items and sort the Database by item decreased support
        buildTree()
            After updating the Database, remaining items will be added into the tree by setting root node as null
        convert()
            to convert the user specified value


    **Methods to execute code on terminal**
    --------------------------------------------
   .. code-block:: console


       Format:


       (.venv) $ python3 parallelPFPGrowth.py <inputFile> <outputFile> <minSup> <maxPer> <noWorker>

       Example usage :

       (.venv) $ python3 parallelPFPGrowth.py sampleTDB.txt patterns.txt 0.3 0.4 5


               .. note:: minSup will be considered in percentage of database transactions


    **Importing this algorithm into a python program**
    -----------------------------------------------------
    .. code-block:: python

                from PAMI.periodicFrequentPattern.basic import parallelPFPGrowth as alg

                obj = alg.parallelPFPGrowth(iFile, minSup, maxPer)

                obj.startMine()

                periodicFrequentPatterns = obj.getPatterns()

                print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))

                obj.savePatterns(oFile)

                Df = obj.getPatternsAsDataFrame()

                memUSS = obj.getMemoryUSS()

                print("Total Memory in USS:", memUSS)

                memRSS = obj.getMemoryRSS()

                print("Total Memory in RSS", memRSS)

                run = obj.getRuntime()

                print("Total ExecutionTime in seconds:", run)

    """
    __startTime = float()
    __endTime = float()
    _minSup = str()
    _maxPer = str()
    _numWorkers = str()
    __finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    __memoryUSS = float()
    __memoryRSS = float()
    __Database = []
    __mapSupport = {}
    __lno = 0
    # __tree = _Tree()
    __rank = {}
    __rankDup = {}
    _numTrans = str()
    __tarunpat = {}

    def __init__(self, iFile, minSup, maxPer, numWorker, sep='\t'):
        super().__init__(iFile, minSup, maxPer, numWorker, sep)

    def func1(self, ps1, tid):
        """
        Add the tid to the set

        :param ps1: set
        :param tid: timestamp of a database
        return: set

        """
        ps1.add(tid)
        return ps1

    def func2(self, ps1, ps2):
        """

        Union of two sets

        :param ps1: set.
        :param ps2: set
        return: set

        """

        ps1 |= ps2
        return ps1

    def func3(self, tids, endts):
        """

        Calculate the periodicity of a transaction

        :param tids: timestamps of a database
        return: periodicity
        """
        # print(tids)
        z = sorted(tids)
        # print(maxPer)
        cur = 0
        for i in z:
            if i - cur > self._maxPer:
                return -1
            cur = i
        if endts - cur > self._maxPer:
            return -1
        else:
            return len(z)

    def getFrequentItems(self, data):
        """
        Get the frequent items from the database

        :param data: database
        return: frequent items
        """
        singleItems = data.flatMap(lambda x: [(x[i], x[0]) for i in range(1, len(x))])
        ps = set()
        freqItems = singleItems.aggregateByKey(ps, lambda ps1, tid: self.func1(ps1, tid),
                                               lambda tuple1, tuple2: self.func2(tuple1, tuple2))
        # print(freqItems.take(10))
        freqItems = freqItems.map(lambda x: (x[0], self.func3(x[1], self._numTrans.value)))
        # print(freqItems.take(10))
        perFreqItems = [x for (x, y) in
                        sorted(freqItems.filter(lambda c: c[1] >= self._minSup).collect(), key=lambda x: -x[1])]
        # print(perFreqItems)
        return perFreqItems

    def getFrequentItemsets(self, data, freqItems):
        """
        Get the frequent itemsets from the database

        :param data: database
        :param freqItems: frequent items
        return: frequent itemsets

        """
        rank = dict([(index, item) for (item, index) in enumerate(self._perFreqItems)])
        numPartitions = data.getNumPartitions()
        workByPartition = data.flatMap(
            lambda basket: self.genCondTransactions(basket[0], basket[1:], rank, numPartitions))
        emptyTree = PFPTree()
        forest = workByPartition.aggregateByKey(emptyTree,
                                                lambda tree, transaction: tree.add(transaction[0], [transaction[1]], 1),
                                                lambda tree1, tree2: tree1.merge(tree2))
        itemsets = forest.flatMap(
            lambda partId_bonsai: partId_bonsai[1].extract(self._minSup, self._maxPer, self._numTrans.value,
                                                           lambda x: self.getPartitionId(x, numPartitions) ==
                                                                     partId_bonsai[0]))
        frequentItemsets = itemsets.map(
            lambda ranks_count: ([self._perFreqItems[z] for z in ranks_count[0]], ranks_count[1]))

        ### TARUN MODIFICATION ###
        frequentItemsets_list = frequentItemsets.collect()

        for itemset, count in frequentItemsets_list:
            string = "\t".join([str(x) for x in itemset])
            self.__tarunpat[string] = count
        ##########################

        return frequentItemsets

    def genCondTransactions(self, tid, basket, rank, nPartitions):
        """
        Get the conditional transactions from the database

        :param tid: timestamp of a database
        :param basket: basket of a database
        :param rank: rank of a database
        :param nPartitions: number of partitions
        """
        filtered = [rank[int(x)] for x in basket if int(x) in rank.keys()]
        filtered = sorted(filtered)
        output = {}
        for i in range(len(filtered) - 1, -1, -1):
            item = filtered[i]
            partition = self.getPartitionId(item, nPartitions)
            if partition not in output.keys():
                output[partition] = [filtered[:i + 1], tid]
        return [x for x in output.items()]

    def getPartitionId(self, key, nPartitions):
        """
        Get the partition id

        :param key: key of a database
        :param nPartitions: number of partitions
        return: partition id

        """
        return key % nPartitions

    def __convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.__Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.__Database) * value)
            else:
                value = int(value)
        return value

    @deprecated("It is recommended to use mine() instead of startMine() for mining process")
    def startMine(self):
        """
        Start the mining process

        """
        self.__startTime = _ab._time.time()

        APP_NAME = "parallelPFPGrowth"
        conf = _ab.SparkConf().setAppName(APP_NAME)
        # conf = conf.setMaster("local[*]")
        sc = _ab.SparkContext(conf=conf).getOrCreate()
        # sc = SparkContext.getOrCreate();
        data = sc.textFile(self._iFile, minPartitions=self._numWorkers).map(
            lambda x: [int(y) for y in x.strip().split(self._sep)])
        # data = sc.textFile(finput).map(lambda x: [int(y) for y in x.strip().split(' ')])
        data.cache()
        # minSupport = data.count() * threshold/100
        # maxPer = data.count() * periodicity_threshold/100
        self._minSup = self.__convert(self._minSup)
        self._maxPer = self.__convert(self._maxPer)
        self._numTrans = sc.broadcast(data.count())
        self._perFreqItems = self.getFrequentItems(data)
        freqItemsets = self.getFrequentItemsets(data, self._perFreqItems)
        self.__finalPatterns = self.__tarunpat
        sc.stop()
        self.__endTime = _ab._time.time()
        self.__memoryUSS = float()
        self.__memoryRSS = float()
        process = _ab._psutil.Process(_ab._os.getpid())
        self.__memoryUSS = process.memory_full_info().uss
        self.__memoryRSS = process.memory_info().rss

    def Mine(self):
        """
        Start the mining process

        """
        self.__startTime = _ab._time.time()

        APP_NAME = "parallelPFPGrowth"
        conf = _ab.SparkConf().setAppName(APP_NAME)
        # conf = conf.setMaster("local[*]")
        sc = _ab.SparkContext(conf=conf).getOrCreate()
        # sc = SparkContext.getOrCreate();
        data = sc.textFile(self._iFile, minPartitions=self._numWorkers).map(
            lambda x: [int(y) for y in x.strip().split(self._sep)])
        # data = sc.textFile(finput).map(lambda x: [int(y) for y in x.strip().split(' ')])
        data.cache()
        # minSupport = data.count() * threshold/100
        # maxPer = data.count() * periodicity_threshold/100
        self._minSup = self.__convert(self._minSup)
        self._maxPer = self.__convert(self._maxPer)
        self._numTrans = sc.broadcast(data.count())
        self._perFreqItems = self.getFrequentItems(data)
        freqItemsets = self.getFrequentItemsets(data, self._perFreqItems)
        self.__finalPatterns = self.__tarunpat
        sc.stop()
        self.__endTime = _ab._time.time()
        self.__memoryUSS = float()
        self.__memoryRSS = float()
        process = _ab._psutil.Process(_ab._os.getpid())
        self.__memoryUSS = process.memory_full_info().uss
        self.__memoryRSS = process.memory_info().rss


    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.__memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.__memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.__endTime - self.__startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self.__tarunpat.items():
            data.append([a.replace('\t', ' '), b])
            dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataframe

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        # print(self.getFrequentItems())
        for x, y in self.__tarunpat.items():
            # print(x,y)
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.__tarunpat

    def printResults(self):
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 6 or len(_ab._sys.argv) == 7:
        if len(_ab._sys.argv) == 7:
            _ap = parallelPFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5],
                                    _ab._sys.argv[6])
        if len(_ab._sys.argv) == 4:
            _ap = parallelPFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        _ap.startMine()
        print("Total number of Frequent Patterns:", _ab.getPatterns())
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        _ap = parallelPFPGrowth('Temporal_T10I4D100K.csv', 500, 5000, 5, '\t')
        _ap.startMine()
        # print("Total number of Frequent Patterns:", len( _ab.getPatterns()))
        # _ap.save(_ab._sys.argv[2])
        _ap.printResults()
        # print("Total Memory in USS:", _ap.getMemoryUSS())
        # print("Total Memory in RSS", _ap.getMemoryRSS())
        # print("Total ExecutionTime in ms:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")


# from PAMI.subgraphMining.topK import tkg as alg

# obj = alg.TKG(iFile, k)

# obj.mine()

# frequentGraphs = obj.getKSubgraphs()

# memUSS = obj.getMemoryUSS()

# obj.save(oFile)

# print("Total Memory in USS:", memUSS)

# memRSS = obj.getMemoryRSS()

# print("Total Memory in RSS", memRSS)

# run = obj.getRuntime()

# print("Total ExecutionTime in seconds:", run)

# minSup = obj.getMinSupport()

# print("Minimum support:", minSup)


from PAMI.subgraphMining.topK import abstract as _ab


class TKG(_ab._TKG):
    ELIMINATE_INFREQUENT_VERTICES = True
    ELIMINATE_INFREQUENT_VERTEX_PAIRS = True
    ELIMINATE_INFREQUENT_EDGE_LABELS = True
    EDGE_COUNT_PRUNING = True
    DYNAMIC_SEARCH = True
    THREADED_DYNAMIC_SEARCH = True

    def __init__(self, iFile, k, maxNumberOfEdges=float('inf'), outputSingleVertices=True, outputGraphIds=False):
        self.iFile = iFile
        self.k = k
        self.outputGraphIds = outputGraphIds
        self.outputSingleVertices = outputSingleVertices
        self.maxNumberOfEdges = maxNumberOfEdges
        self.graphCount = 0
        self.patternCount = 0
        self.frequentVertexLabels = []
        self.infrequentVerticesRemovedCount = 0
        self.infrequentVertexPairsRemovedCount = 0
        self.skipStrategyCount = 0
        self.threadCount = 1
        self.edgeRemovedByLabel = 0
        self.eliminatedWithMaxSize = 0
        self.emptyGraphsRemoved = 0
        self.pruneByEdgeCount = 0


    def mine(self):
        """
        This Python function starts a mining process on a graph database, calculates runtime, pattern count,
        and memory usage metrics.
        """
        if self.maxNumberOfEdges <= 0:
            return

        self.kSubgraphs = _ab.PriorityQueue()
        self.candidates = _ab.PriorityQueue()
        
        self.runtime = 0

        t1 = _ab.time.time()
        graphDb = self.readGraphs(self.iFile)
        self.minSup = 1

        self.gSpan(graphDb, self.outputSingleVertices)

        t2 = _ab.time.time()
        self.runtime = t2 - t1
        self.patternCount = self.getQueueSize(self.kSubgraphs)

        process = _ab._psutil.Process(_ab._os.getpid())

        self._memoryUSS = float()

        self._memoryRSS = float()

        self._memoryUSS = process.memory_full_info().uss

        self._memoryRSS = process.memory_info().rss



    def readGraphs(self, path):
        """
        The `readGraphs` function reads graph data from a file and constructs a list of graphs with vertices
        and edges.
        
        :param path: This method reads the graph data from the specified file and constructs a list of graphs 
        represented by vertices and edges
        :return: The `readGraphs` method returns a list of `_ab.Graph` objects, which represent graphs read
        from the file.
        """
        with open(path, 'r') as br:
            graphDatabase = []
            vMap = {}
            gId = None

            for line in br:
                line = line.strip()
                if line.startswith("t"):
                    if vMap:
                        graphDatabase.append(_ab.Graph(gId, vMap))
                        vMap = {}
                    gId = int(line.split()[2])
                elif line.startswith("v"):
                    items = line.split()
                    vId, vLabel = int(items[1]), int(items[2])
                    vMap[vId] = _ab.Vertex(vId, vLabel)
                elif line.startswith("e"):
                    items = line.split()
                    v1, v2, eLabel = int(items[1]), int(items[2]), int(items[3])
                    edge = _ab.Edge(v1, v2, eLabel)
                    vMap[v1].addEdge(edge)
                    vMap[v2].addEdge(edge)

            if vMap:
                graphDatabase.append(_ab.Graph(gId, vMap))

        self.graphCount = len(graphDatabase)
        return graphDatabase

    def save(self, oFile):
        """
        The `save` function writes subgraph information to a file in a specific format.
        
        :param oFile: The `oFile` parameter in the `save` method is the file path where the output will be
        saved. This method writes the subgraphs information to the specified file in a specific format
        """
        subgraphsList = self.getSubgraphsList()

        with open(oFile, 'w') as bw:
            for i, subgraph in enumerate(subgraphsList):
                sb = []
                dfsCode = subgraph.dfsCode

                sb.append(f"t # {i} * {subgraph.support}\n")
                if len(dfsCode.eeList) == 1:
                    ee = dfsCode.eeList[0]
                    sb.append(f"v 0 {ee.vLabel1}\n")
                    if ee.edgeLabel != -1:
                        sb.append(f"v 1 {ee.vLabel2}\n")
                        sb.append(f"e 0 1 {ee.edgeLabel}\n")
                else:
                    vLabels = dfsCode.getAllVLabels()
                    for j, vLabel in enumerate(vLabels):
                        sb.append(f"v {j} {vLabel}\n")
                    for ee in dfsCode.eeList:
                        sb.append(f"e {ee.v1} {ee.v2} {ee.edgeLabel}\n")

                if self.outputGraphIds:
                    sb.append("x " + " ".join(str(id) for id in subgraph.setOfGraphsIds))
                sb.append("\n\n")
                bw.write("".join(sb))


    def savePattern(self, subgraph):        
        # previousMinSup = self.minSup

        self.kSubgraphs.put(subgraph)
        if self.kSubgraphs.qsize() > self.k:
            while self.kSubgraphs.qsize() > self.k:
                lower = self.kSubgraphs.get()

                if lower.support > self.minSup:
                    self.minSup = lower.support


    def getQueueSize(self, queue):
        size = 0
        tempQueue = _ab.PriorityQueue()
        
        while not queue.empty():
            item = queue.get()
            tempQueue.put(item)
            size += 1
        
        while not tempQueue.empty():
            queue.put(tempQueue.get())
        
        return size

    def subgraphIsomorphisms(self, c, g):
        isoms = []
        startLabel = c.getEeList()[0].vLabel1
        for vId in g.findAllWithLabel(startLabel):
            isoms.append({0: vId})
        
        for ee in c.getEeList():
            v1, v2, v2Label, eLabel = ee.v1, ee.v2, ee.vLabel2, ee.edgeLabel
            updateIsoms = []
            for iso in isoms:
                mappedV1 = iso[v1]
                if v1 < v2:
                    mappedVertices = set(iso.values())
                    for mappedV2 in g.getAllNeighbors(mappedV1):
                        if v2Label == mappedV2.getLabel() and mappedV2.getId() not in mappedVertices and eLabel == g.getEdgeLabel(mappedV1, mappedV2.getId()):
                            tempIso = iso.copy()
                            tempIso[v2] = mappedV2.getId()
                            updateIsoms.append(tempIso)
                else:
                    mappedV2 = iso[v2]
                    if g.isNeighboring(mappedV1, mappedV2) and eLabel == g.getEdgeLabel(mappedV1, mappedV2):
                        updateIsoms.append(iso)
            isoms = updateIsoms
        return isoms



    def rightMostPathExtensionsFromSingle(self, c, g):
        gid = g.getId()
        extensions = {}
    
        if c.isEmpty():
            for vertex in g.vertices:
                for e in vertex.getEdgeList():
                    v1Label = g.getVLabel(e.v1)
                    v2Label = g.getVLabel(e.v2)
                    if v1Label < v2Label:
                        ee1 = _ab.ExtendedEdge(0, 1, v1Label, v2Label, e.getEdgeLabel())
                    else:
                        ee1 = _ab.ExtendedEdge(0, 1, v2Label, v1Label, e.getEdgeLabel())
                    extensions.setdefault(ee1, set()).add(gid)
        else:
            rightMost = c.getRightMost()
            isoms = self.subgraphIsomorphisms(c, g)
            for iso in isoms:
                invertedIsom = {v: k for k, v in iso.items()}
                mappedRm = iso[rightMost]
                mappedRmLabel = g.getVLabel(mappedRm)
                for x in g.getAllNeighbors(mappedRm):
                    invertedX = invertedIsom.get(x.getId())
                    if invertedX is not None and c.onRightMostPath(invertedX) and not c.containEdge(rightMost, invertedX):
                        ee = _ab.ExtendedEdge(rightMost, invertedX, mappedRmLabel, x.getLabel(), g.getEdgeLabel(mappedRm, x.getId()))
                        extensions.setdefault(ee, set()).add(gid)
                mappedVertices = set(iso.values())
                for v in c.getRightMostPath():
                    mappedV = iso[v]
                    mappedVLabel = g.getVLabel(mappedV)
                    for x in g.getAllNeighbors(mappedV):
                        if x.getId() not in mappedVertices:
                            ee = _ab.ExtendedEdge(v, rightMost + 1, mappedVLabel, x.getLabel(), g.getEdgeLabel(mappedV, x.getId()))
                            extensions.setdefault(ee, set()).add(gid)
        return extensions

    def rightMostPathExtensions(self, c, graphDB, graphIds):
        extensions = {}
    
        if c.isEmpty():
            for graphId in graphIds:
                g = graphDB[graphId]
                if self.EDGE_COUNT_PRUNING and c.size >= g.getEdgeCount():
                    self.pruneByEdgeCount += 1
                    continue
                for vertex in g.vertices:
                    for e in vertex.getEdgeList():
                        v1Label = g.getVLabel(e.v1)
                        v2Label = g.getVLabel(e.v2)
                        if v1Label < v2Label:
                            ee1 = _ab.ExtendedEdge(0, 1, v1Label, v2Label, e.getEdgeLabel())
                        else:
                            ee1 = _ab.ExtendedEdge(0, 1, v2Label, v1Label, e.getEdgeLabel())
                        extensions.setdefault(ee1, set()).add(graphId)
        else:
            rightMost = c.getRightMost()
            for graphId in graphIds:
                g = graphDB[graphId]
                if self.EDGE_COUNT_PRUNING and c.size >= g.getEdgeCount():
                    self.pruneByEdgeCount += 1
                    continue
                isoms = self.subgraphIsomorphisms(c, g)
                for isom in isoms:
                    invertedIsom = {v: k for k, v in isom.items()}
                    mappedRm = isom[rightMost]
                    mappedRmLabel = g.getVLabel(mappedRm)
                    for x in g.getAllNeighbors(mappedRm):
                        invertedX = invertedIsom.get(x.getId())
                        if invertedX is not None and c.onRightMostPath(invertedX) and not c.containEdge(rightMost, invertedX):
                            ee = _ab.ExtendedEdge(rightMost, invertedX, mappedRmLabel, x.getLabel(), g.getEdgeLabel(mappedRm, x.getId()))
                            extensions.setdefault(ee, set()).add(g.getId())
                    mappedVertices = set(isom.values())
                    for v in c.getRightMostPath():
                        mappedV = isom[v]
                        mappedVLabel = g.getVLabel(mappedV)
                        for x in g.getAllNeighbors(mappedV):
                            if x.getId() not in mappedVertices:
                                ee = _ab.ExtendedEdge(v, rightMost + 1, mappedVLabel, x.getLabel(), g.getEdgeLabel(mappedV, x.getId()))
                                extensions.setdefault(ee, set()).add(g.getId())
        return extensions



    def gSpan(self, graphDB, outputFrequentVertices):
        if outputFrequentVertices or self.ELIMINATE_INFREQUENT_VERTICES:
            self.findAllOnlyOneVertex(graphDB, outputFrequentVertices)
        
        for g in graphDB:
            g.precalculateVertexList()
    
        if self.ELIMINATE_INFREQUENT_VERTEX_PAIRS or self.ELIMINATE_INFREQUENT_EDGE_LABELS:
            self.removeInfrequentVertexPairs(graphDB)
    
        graphIds = set()
        for i, g in enumerate(graphDB):
            if g.vertices and len(g.vertices) != 0:
                if self.infrequentVerticesRemovedCount > 0:
                    g.precalculateVertexList()
    
                graphIds.add(i)
                g.precalculateVertexNeighbors()
                g.precalculateLabelsToVertices()
            else:
                self.emptyGraphsRemoved += 1
    
        if not outputFrequentVertices or self.frequentVertexLabels:
            if self.DYNAMIC_SEARCH:
                self.gspanDynamicDFS(_ab.DfsCode(), graphDB, graphIds)
                
                if self.THREADED_DYNAMIC_SEARCH:
                    self.startThreads(graphDB, self.candidates, self.minSup)

                else:
                    while self.candidates:
                        candidate = self.candidates.pop()
                        if len(candidate.setOfGraphsIds) < self.minSup:
                            continue
                        self.gspanDynamicDFS(candidate.dfsCode, graphDB, candidate.setOfGraphsIds)
            else:
                self.gspanDfs(_ab.DfsCode(), graphDB, graphIds)

    def startThreads(self, graphDB, candidates, minSup):
        threads = []
        for _ in range(self.threadCount):
            thread = _ab.DfsThread(graphDB, candidates, minSup, self)
            thread.start()
            threads.append(thread)
    
        for thread in threads:
            thread.join()

    def gspanDfs(self, c: _ab.DfsCode, graphDB, subgraphId):
        if c.size == self.maxNumberOfEdges - 1:
            return
        extensions = self.rightMostPathExtensions(c, graphDB, subgraphId)
        for extension, newGraphIds in extensions.items():
            sup = len(newGraphIds)
            if sup >= self.minSup:
                newC = c.copy()
                newC.add(extension)
    
                if self.isCanonical(newC):
                    self.savePattern(_ab.FrequentSubgraph(newC, newGraphIds, sup))
                    self.gspanDfs(newC, graphDB, newGraphIds)

    
    def gspanDynamicDFS(self, c, graphDB, graphIds):
        if c.size == self.maxNumberOfEdges - 1:
            return

        extensions = self.rightMostPathExtensions(c, graphDB, graphIds)
        for extension, newGraphIds in extensions.items():
            support = len(newGraphIds)

            if support >= self.minSup:
                newC = c.copy()
                newC.add(extension)
                if self.isCanonical(newC):
                    subgraph = _ab.FrequentSubgraph(newC, newGraphIds, support)
                    self.savePattern(subgraph)
                    self.registerAsCandidate(subgraph)
    
    def registerAsCandidate(self, subgraph):
        self.candidates.put((-subgraph.support, subgraph))

    
    def isCanonical(self, c: _ab.DfsCode):
        canC = _ab.DfsCode()
        for i in range(c.size):
            extensions = self.rightMostPathExtensionsFromSingle(canC, _ab.Graph(c))
            minEe = None
            for ee in extensions.keys():
                if minEe is None or ee.smallerThan(minEe):
                    minEe = ee

            if minEe is not None and minEe.smallerThan(c.getAt(i)):
                return False
            
            if minEe is not None:
                canC.add(minEe)
        return True


    class Pair:
        def __init__(self, x, y):
            if x < y:
                self.x = x
                self.y = y
            else:
                self.x = y
                self.y = x

        def __eq__(self, other):
            if isinstance(other, TKG.Pair):
                return self.x == other.x and self.y == other.y
            return False

        def __hash__(self):
            return self.x + 100 * self.y


    def findAllOnlyOneVertex(self, graphDB, outputFrequentVertices):
        self.frequentVertexLabels = []
        labelM = {} 
        for g in graphDB:
            for v in g.getNonPrecalculatedAllVertices():
                if v.getEdgeList():
                    vLabel = v.getLabel()
                    labelM.setdefault(vLabel, set()).add(g.getId())
        for label, tempSupG in labelM.items():                
            sup = len(tempSupG)
            if sup >= self.minSup:
                self.frequentVertexLabels.append(label)
                if outputFrequentVertices:
                    tempD = _ab.DfsCode()
                    tempD.add(_ab.ExtendedEdge(0, 0, label, label, -1))
                    self.savePattern(_ab.FrequentSubgraph(tempD, tempSupG, sup))
            elif TKG.ELIMINATE_INFREQUENT_VERTICES:
                for graphId in tempSupG:
                    g = graphDB[graphId]
                    g.removeInfrequentLabel(label)
                    self.infrequentVerticesRemovedCount += 1

    def removeInfrequentVertexPairs(self, graphDB):
        if TKG.ELIMINATE_INFREQUENT_EDGE_LABELS:
            matrix = _ab.SparseTriangularMatrix()
            alreadySeenPair = set()

        if TKG.ELIMINATE_INFREQUENT_EDGE_LABELS:
            mapEdgeLabelToSupport = {}
            alreadySeenEdgeLabel = set()

        for g in graphDB:
            vertices = g.getAllVertices()

            for v1 in vertices:
                labelV1 = v1.getLabel()

                for edge in v1.getEdgeList():
                    v2 = edge.another(v1.getId())
                    labelV2 = g.getVLabel(v2)

                    if TKG.ELIMINATE_INFREQUENT_EDGE_LABELS:
                        pair = self.Pair(labelV1, labelV2)
                        if pair not in alreadySeenPair:
                            matrix.incrementCount(labelV1, labelV2)
                            alreadySeenPair.add(pair)

                    if TKG.ELIMINATE_INFREQUENT_EDGE_LABELS:
                        edgeLabel = edge.getEdgeLabel()
                        if edgeLabel not in alreadySeenEdgeLabel:
                            alreadySeenEdgeLabel.add(edgeLabel)
                            edgeSupport = mapEdgeLabelToSupport.get(edgeLabel, 0)
                            mapEdgeLabelToSupport[edgeLabel] = edgeSupport + 1

            if TKG.ELIMINATE_INFREQUENT_VERTEX_PAIRS:
                alreadySeenPair.clear()
            if TKG.ELIMINATE_INFREQUENT_EDGE_LABELS:
                alreadySeenEdgeLabel.clear()

        if TKG.ELIMINATE_INFREQUENT_VERTEX_PAIRS:
            matrix.removeInfrequentEntriesFromMatrix(self.minSup)

        if TKG.ELIMINATE_INFREQUENT_VERTEX_PAIRS or TKG.ELIMINATE_INFREQUENT_EDGE_LABELS:
            for g in graphDB:
                vertices = g.getAllVertices()

                for v1 in vertices:
                    iterEdges = iter(v1.getEdgeList())
                    for edge in iterEdges:
                        v2 = edge.another(v1.getId())
                        labelV2 = g.getVLabel(v2)
                        count = matrix.getSupportForItems(v1.getLabel(), labelV2)

                        if TKG.ELIMINATE_INFREQUENT_VERTEX_PAIRS and count < self.minSup:
                            v1.removeEdge(edge)
                            self.infrequentVertexPairsRemovedCount += 1

                        elif TKG.ELIMINATE_INFREQUENT_EDGE_LABELS and \
                                mapEdgeLabelToSupport.get(edge.getEdgeLabel(), 0) < self.minSup:
                            v1.removeEdge(edge)
                            self.edgeRemovedByLabel += 1

    def getMemoryRSS(self):
        return self._memoryRSS

    def getMemoryUSS(self):
        return self._memoryUSS

    def getRuntime(self):
        return self.runtime
    
    def getMinSupport(self):
        return self.minSup
    
    def getKSubgraphs(self):
        """ Return the formatted subgraphs as a single string with correct formatting and newlines. """
        subgraphsList = self.getSubgraphsList()  
        sb = [] 
        for i, subgraph in enumerate(subgraphsList):
            subgraphDescription = [f"t # {i} * {subgraph.support}"]  
            dfsCode = subgraph.dfsCode
            if len(dfsCode.eeList) == 1:
                ee = dfsCode.eeList[0]
                subgraphDescription.append(f"v 0 {ee.vLabel1}")
                if ee.edgeLabel != -1:
                    subgraphDescription.append(f"v 1 {ee.vLabel2}")
                    subgraphDescription.append(f"e 0 1 {ee.edgeLabel}")
            else:
                vLabels = dfsCode.getAllVLabels()
                for j, vLabel in enumerate(vLabels):
                    subgraphDescription.append(f"v {j} {vLabel}")
                for ee in dfsCode.eeList:
                    subgraphDescription.append(f"e {ee.v1} {ee.v2} {ee.edgeLabel}")

            # Include graph IDs if the feature is enabled
            if self.outputGraphIds and subgraph.setOfGraphsIds:
                subgraphDescription.append("x " + " ".join(str(id) for id in subgraph.setOfGraphsIds))
            sb.append('\n'.join(subgraphDescription)) 
        return '\n\n'.join(sb)  



    def getSubgraphsList(self):
        """Creates a copy of the queue's contents without emptying the original queue."""
        subgraphsList = list(self.kSubgraphs.queue)
        subgraphsList.sort(key=lambda sg: sg.support, reverse=True)
        return subgraphsList


    




# MultipleTimeSeriesFuzzyDatabaseStats is class to get stats of multiple time series fuzzy database.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.dbStats import MultipleTimeSeriesFuzzyDatabaseStats as db
#
#             obj = db.MultipleTimeSeriesDatabaseStats(iFile, "\t")
#
#             obj.run()
#
#             obj.printStats()
#
#             obj.save(oFile)
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import statistics
import pandas as pd
import validators
import numpy as np
from urllib.request import urlopen
import sys
import PAMI.extras.graph.plotLineGraphFromDictionary as plt


class MultipleTimeSeriesFuzzyDatabaseStats:
    """
    :Description:  MultipleTimeSeriesDatabaseStats is class to get stats of multiple time series fuzzy database.

    :Attributes:

        :param inputFile: file :
            input file path
        :param sep: str
            separator in file. Default is tab space.

    :Methods:

        run()
            execute readDatabase function
        readDatabase()
            read database from input file
        getDatabaseSize()
            get the size of database
        getTotalNumberOfItems()
            get the total number of items in a database
        getMinimumTransactionLength()
            get the minimum transaction length
        getAverageTransactionLength()
            get the average transaction length. It is sum of all transaction length divided by database length.
        getMaximumTransactionLength()
            get the maximum transaction length
        getStandardDeviationTransactionLength()
            get the standard deviation of transaction length
        convertDataIntoMatrix()
            Convert the database into matrix form to calculate the sparsity and density of a database
        getSparsity()
            get sparsity value of database
        getDensity()
            get density value of database
        getSortedListOfItemFrequencies()
            get sorted list of item frequencies
        getSortedListOfTransactionLength()
            get sorted list of transaction length
        save(data, outputFile)
            store data into outputFile
        printStats()
            To print all the stats of the database
        plotGraphs()
            To plot all the graphs of frequency disctribution of items and transaction length distribution in database
   

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.dbStats import MultipleTimeSeriesFuzzyDatabaseStats as db

            obj = db.MultipleTimeSeriesFuzzyDatabaseStats(iFile, "\t")

            obj.run()

            obj.save(oFile)

            obj.printStats()

    """

    def __init__(self, inputFile: str, sep: str='\t'):
        """
        :param inputFile: input file name or path
        :type inputFile: str
        :param sep: separator
        :type sep: str
        """
        self.inputFile = inputFile
        self.lengthList = []
        self.sep = sep
        self.database = {}
        self.itemFrequencies = {}

    def run(self) -> None:
        self.readDatabase()

    def readDatabase(self) -> None:
        """
        read database from input file and store into database and size of each transaction.
        """
        self._transactions, self._fuzzyValues, self._Database, self._ts = [], [], [], []
        numberOfTransaction = 0
        if isinstance(self.inputFile, pd.DataFrame):
            if self.inputFile.empty:
                print("its empty..")
            i = self.inputFile.columns.values.tolist()
            if 'tid' in i and 'Transactions' in i:
                self.database = self.inputFile.set_index('tid').T.to_dict(orient='records')[0]
            if 'tid' in i and 'Patterns' in i:
                self.database = self.inputFile.set_index('tid').T.to_dict(orient='records')[0]
        if isinstance(self.inputFile, str):
            if validators.url(self.inputFile):
                data = urlopen(self.inputFile)
                for line in data:
                    numberOfTransaction += 1
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self.sep)]
                    temp = [x for x in temp if x]
                    self.database[numberOfTransaction] = temp
            else:
                try:
                    with open(self.inputFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.strip()
                            parts = line.split(":")
                            numberOfTransaction += 1
                            parts[0] = parts[0].strip()
                            parts[1] = parts[1].strip()
                            parts[2] = parts[2].strip()
                            times = parts[0].split(self.sep)
                            items = parts[1].split(self.sep)
                            quantities = parts[2].split(self.sep)
                            #print(times, items, quantities)
                            _time = [x for x in times if x]
                            items = [x for x in items if x]
                            quantities = [float(x) for x in quantities if x]
                            tempList = []
                            for k in range(len(_time)):
                                ite = "(" + _time[k] + "," + items[k] + ")"
                                tempList.append(ite)
                            self._ts.append([x for x in times])
                            self._transactions.append([x for x in tempList])
                            self._fuzzyValues.append([x for x in quantities])
                            self.database[numberOfTransaction] = tempList
                except IOError:
                    print("File Not Found")
                    quit()
        self.lengthList = [len(s) for s in self._transactions]

    def getDatabaseSize(self) -> int:
        """
        get the size of database
        :return: dataset size
        :rtype: int
        """
        return len(self.database)

    def getTotalNumberOfItems(self) -> int:
        """
        get the number of items in database.
        :return: number of items
        :rtype: int
        """
        return len(self.getSortedListOfItemFrequencies())

    def getMinimumTransactionLength(self) -> int:
        """
        get the minimum transaction length
        :return: minimum transaction length
        :rtype: int
        """
        return min(self.lengthList)

    def getAverageTransactionLength(self) -> float:
        """
        get the average transaction length. It is sum of all transaction length divided by database length.
        :return: average transaction length
        :rtype: float
        """
        totalLength = sum(self.lengthList)
        return totalLength / len(self.database)

    def getMaximumTransactionLength(self) -> int:
        """
        get the maximum transaction length
        :return: maximum transaction length
        :rtype: int
        """
        return max(self.lengthList)

    def getStandardDeviationTransactionLength(self) -> float:
        """
        get the standard deviation transaction length
        :return: standard deviation transaction length
        :rtype: float
        """
        return statistics.pstdev(self.lengthList)

    def getVarianceTransactionLength(self) -> float:
        """
        get the variance transaction length
        :return: variance transaction length
        :rtype: float
        """
        return statistics.variance(self.lengthList)

    def getNumberOfItems(self) -> int:
        """
        get the number of items in database.
        :return: number of items
        :rtype: int
        """
        return len(self.getSortedListOfItemFrequencies())

    def convertDataIntoMatrix(self) -> np.ndarray:
        singleItems = self.getSortedListOfItemFrequencies()
        # big_array = np.zeros((self.getDatabaseSize(), len(self.getSortedListOfItemFrequencies())))
        itemsets = {}
        for i in self.database:
            for item in singleItems:
                if item in itemsets:
                    if item in self.database[i]:
                        itemsets[item].append(1)
                    else:
                        itemsets[item].append(0)
                else:
                    if item in self.database[i]:
                        itemsets[item] = [1]
                    else:
                        itemsets[item] = [0]
        # new = pd.DataFrame.from_dict(itemsets)
        data = list(itemsets.values())
        an_array = np.array(data)
        return an_array

    def getSparsity(self) -> float:
        """
        get the sparsity of database. sparsity is percentage of 0 of database.
        :return: database sparsity
        :rtype: float
        """
        big_array = self.convertDataIntoMatrix()
        n_zeros = np.count_nonzero(big_array == 0)
        return (n_zeros / big_array.size)

    def getDensity(self) -> float:
        """
        get the sparsity of database. sparsity is percentage of 0 of database.
        :return: database sparsity
        :rtype: float
        """
        big_array = self.convertDataIntoMatrix()
        n_zeros = np.count_nonzero(big_array != 0)
        return (n_zeros / big_array.size)

    def getSortedListOfItemFrequencies(self) -> dict:
        """
        get sorted list of item frequencies
        :return: item frequencies
        :rtype: dict
        """
        itemFrequencies = {}
        for line in range(len(self._transactions)):
            times = self._ts[line]
            items = self._transactions[line]
            quantities = self._fuzzyValues[line]
            for i in range(0, len(items)):
                item = items[i]
                if item in itemFrequencies:
                    itemFrequencies[item] += quantities[i]
                else:
                    itemFrequencies[item] = quantities[i]
        self.itemFrequencies = {k: v for k, v in sorted(itemFrequencies.items(), key=lambda x: x[1], reverse=True)}
        return self.itemFrequencies
    
    def getFrequenciesInRange(self) -> dict:
        fre = self.getSortedListOfItemFrequencies()
        rangeFrequencies = {}
        maximum = max([i for i in fre.values()])
        values = [int(i*maximum/6) for i in range(1,6)]
        va = len({key: val for key, val in fre.items() if val > 0 and val < values[0]})
        rangeFrequencies[va] = values[0]
        for i in range(1,len(values)):
            va = len({key: val for key, val in fre.items() if val < values[i] and val > values[i-1]})
            rangeFrequencies[va] = values[i]
        return rangeFrequencies

    def getTransanctionalLengthDistribution(self) -> dict:
        """
        get transaction length
        :return: transactional length
        :rtype: dict
        """
        transactionLength = {}
        for length in self.lengthList:
            transactionLength[length] = transactionLength.get(length, 0)
            transactionLength[length] += 1
        return {k: v for k, v in sorted(transactionLength.items(), key=lambda x: x[0])}

    def save(self, data: dict, outputFile: str) -> None:
        """
        store data into outputFile
        :param data: input data
        :type data: dict
        :param outputFile: output file name or path to store
        :type outputFile: str
        :return: None
        """
        with open(outputFile, 'w') as f:
            for key, value in data.items():
                f.write(f'{key}\t{value}\n')
                   
    def printStats(self) -> None:
        print(f'Database size (total no of transactions) : {self.getDatabaseSize()}')
        print(f'Number of items : {self.getNumberOfItems()}')
        print(f'Minimum Transaction Size : {self.getMinimumTransactionLength()}')
        print(f'Average Transaction Size : {self.getAverageTransactionLength()}')
        print(f'Maximum Transaction Size : {self.getMaximumTransactionLength()}')
        print(f'Standard Deviation Transaction Size : {self.getStandardDeviationTransactionLength()}')
        print(f'Variance in Transaction Sizes : {self.getVarianceTransactionLength()}')
        print(f'Sparsity : {self.getSparsity()}')
  
    def plotGraphs(self) -> None:
        # itemFrequencies = self.getFrequenciesInRange()
        transactionLength = self.getTransanctionalLengthDistribution()
        plt.plotLineGraphFromDictionary(self.itemFrequencies, 100, 0, 'Frequency', 'No of items', 'frequency')
        plt.plotLineGraphFromDictionary(transactionLength, 100, 0, 'transaction length', 'transaction length', 'frequency')


if __name__ == '__main__':
    import PAMI.extras.graph.plotLineGraphFromDictionary as plt
    import pandas as pd
    obj = MultipleTimeSeriesFuzzyDatabaseStats(sys.argv[1])
    obj.run()
    obj.printStats()
    obj.plotGraphs()
